#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

from itertools import chain


def multidigraph_edge_data_iterator(graph):
    for _, _, _, edge_data in graph.edges(data=True, keys=True):
        yield edge_data


def multidigraph_node_data_iterator(graph):
    for _, node_data in graph.nodes(data=True):
        yield node_data


def multidigraph_data_iterator(graph):
    return chain(multidigraph_node_data_iterator(graph), multidigraph_edge_data_iterator(graph))


#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import inspect
import time

from grakn.client import GraknClient

from kglib.kgcn.pipeline.pipeline import pipeline
from kglib.utils.grakn.synthetic.examples.diagnosis.generate import generate_example_graphs
from kglib.utils.grakn.type.type import get_thing_types, get_role_types
from kglib.utils.graph.iterate import multidigraph_data_iterator
from kglib.utils.graph.query.query_graph import QueryGraph
from kglib.utils.graph.thing.queries_to_graph import build_graph_from_queries

KEYSPACE = "diagnosis"
URI = "localhost:48555"

# Existing elements in the graph are those that pre-exist in the graph, and should be predicted to continue to exist
PREEXISTS = 0

# Candidates are neither present in the input nor in the solution, they are negative samples
CANDIDATE = 1

# Elements to infer are the graph elements whose existence we want to predict to be true, they are positive samples
TO_INFER = 2

# Categorical Attribute types and the values of their categories
CATEGORICAL_ATTRIBUTES = {'name': ['Diabetes Type II', 'Multiple Sclerosis', 'Blurred vision', 'Fatigue', 'Cigarettes',
                                   'Alcohol']}
# Continuous Attribute types and their min and max values
CONTINUOUS_ATTRIBUTES = {'severity': (0, 1), 'age': (7, 80), 'units-per-week': (3, 29)}

TYPES_TO_IGNORE = ['candidate-diagnosis', 'example-id', 'probability-exists', 'probability-non-exists', 'probability-preexists']
ROLES_TO_IGNORE = ['candidate-patient', 'candidate-diagnosed-disease']

# The learner should see candidate relations the same as the ground truth relations, so adjust these candidates to
# look like their ground truth counterparts
TYPES_AND_ROLES_TO_OBFUSCATE = {'candidate-diagnosis': 'diagnosis',
                                'candidate-patient': 'patient',
                                'candidate-diagnosed-disease': 'diagnosed-disease'}


def diagnosis_example(num_graphs=200,
                      num_processing_steps_tr=5,
                      num_processing_steps_ge=5,
                      num_training_iterations=1000,
                      keyspace=KEYSPACE, uri=URI):
    """
    Run the diagnosis example from start to finish, including traceably ingesting predictions back into Grakn

    Args:
        num_graphs: Number of graphs to use for training and testing combined
        num_processing_steps_tr: The number of message-passing steps for training
        num_processing_steps_ge: The number of message-passing steps for testing
        num_training_iterations: The number of training epochs
        keyspace: The name of the keyspace to retrieve example subgraphs from
        uri: The uri of the running Grakn instance

    Returns:
        Final accuracies for training and for testing
    """

    tr_ge_split = int(num_graphs*0.5)

    generate_example_graphs(num_graphs, keyspace=keyspace, uri=uri)

    client = GraknClient(uri=uri)
    session = client.session(keyspace=keyspace)

    graphs = create_concept_graphs(list(range(num_graphs)), session)

    with session.transaction().read() as tx:
        # Change the terminology here onwards from thing -> node and role -> edge
        node_types = get_thing_types(tx)
        [node_types.remove(el) for el in TYPES_TO_IGNORE]

        edge_types = get_role_types(tx)
        [edge_types.remove(el) for el in ROLES_TO_IGNORE]
        print(f'Found node types: {node_types}')
        print(f'Found edge types: {edge_types}')

    ge_graphs, solveds_tr, solveds_ge = pipeline(graphs,
                                                 tr_ge_split,
                                                 node_types,
                                                 edge_types,
                                                 num_processing_steps_tr=num_processing_steps_tr,
                                                 num_processing_steps_ge=num_processing_steps_ge,
                                                 num_training_iterations=num_training_iterations,
                                                 continuous_attributes=CONTINUOUS_ATTRIBUTES,
                                                 categorical_attributes=CATEGORICAL_ATTRIBUTES,
                                                 output_dir=f"./events/{time.time()}/")

    with session.transaction().write() as tx:
        write_predictions_to_grakn(ge_graphs, tx)

    session.close()
    client.close()

    return solveds_tr, solveds_ge


def create_concept_graphs(example_indices, grakn_session):
    """
    Builds an in-memory graph for each example, with an example_id as an anchor for each example subgraph.
    Args:
        example_indices: The values used to anchor the subgraph queries within the entire knowledge graph
        grakn_session: Grakn Session

    Returns:
        In-memory graphs of Grakn subgraphs
    """

    graphs = []
    infer = True

    for example_id in example_indices:
        print(f'Creating graph for example {example_id}')
        graph_query_handles = get_query_handles(example_id)
        with grakn_session.transaction().read() as tx:
            # Build a graph from the queries, samplers, and query graphs
            graph = build_graph_from_queries(graph_query_handles, tx, infer=infer)

        obfuscate_labels(graph, TYPES_AND_ROLES_TO_OBFUSCATE)

        graph.name = example_id
        graphs.append(graph)

    return graphs


def obfuscate_labels(graph, types_and_roles_to_obfuscate):
    # Remove label leakage - change type labels that indicate candidates into non-candidates
    for data in multidigraph_data_iterator(graph):
        for label_to_obfuscate, with_label in types_and_roles_to_obfuscate.items():
            if data['type'] == label_to_obfuscate:
                data.update(type=with_label)
                break


def get_query_handles(example_id):
    """
    Creates an iterable, each element containing a Graql query, a function to sample the answers, and a QueryGraph
    object which must be the Grakn graph representation of the query. This tuple is termed a "query_handle"

    Args:
        example_id: A uniquely identifiable attribute value used to anchor the results of the queries to a specific
                    subgraph

    Returns:
        query handles
    """

    # === Hereditary Feature ===
    hereditary_query = inspect.cleandoc(f'''match
           $p isa person, has example-id {example_id};
           $par isa person;
           $ps(child: $p, parent: $par) isa parentship;
           $diag(patient:$par, diagnosed-disease: $d) isa diagnosis;
           $d isa disease, has name $n;
           get;''')

    vars = p, par, ps, d, diag, n = 'p', 'par', 'ps', 'd', 'diag', 'n'
    hereditary_query_graph = (QueryGraph()
                              .add_vars(vars, PREEXISTS)
                              .add_role_edge(ps, p, 'child', PREEXISTS)
                              .add_role_edge(ps, par, 'parent', PREEXISTS)
                              .add_role_edge(diag, par, 'patient', PREEXISTS)
                              .add_role_edge(diag, d, 'diagnosed-disease', PREEXISTS)
                              .add_has_edge(d, n, PREEXISTS))

    # === Consumption Feature ===
    consumption_query = inspect.cleandoc(f'''match
           $p isa person, has example-id {example_id};
           $s isa substance, has name $n;
           $c(consumer: $p, consumed-substance: $s) isa consumption, 
           has units-per-week $u; get;''')

    vars = p, s, n, c, u = 'p', 's', 'n', 'c', 'u'
    consumption_query_graph = (QueryGraph()
                               .add_vars(vars, PREEXISTS)
                               .add_has_edge(s, n, PREEXISTS)
                               .add_role_edge(c, p, 'consumer', PREEXISTS)
                               .add_role_edge(c, s, 'consumed-substance', PREEXISTS)
                               .add_has_edge(c, u, PREEXISTS))

    # === Age Feature ===
    person_age_query = inspect.cleandoc(f'''match 
            $p isa person, has example-id {example_id}, has age $a; 
            get;''')

    vars = p, a = 'p', 'a'
    person_age_query_graph = (QueryGraph()
                              .add_vars(vars, PREEXISTS)
                              .add_has_edge(p, a, PREEXISTS))

    # === Risk Factors Feature ===
    risk_factor_query = inspect.cleandoc(f'''match 
            $d isa disease; 
            $p isa person, has example-id {example_id}; 
            $r(person-at-risk: $p, risked-disease: $d) isa risk-factor; 
            get;''')

    vars = p, d, r = 'p', 'd', 'r'
    risk_factor_query_graph = (QueryGraph()
                               .add_vars(vars, PREEXISTS)
                               .add_role_edge(r, p, 'person-at-risk', PREEXISTS)
                               .add_role_edge(r, d, 'risked-disease', PREEXISTS))

    # === Symptom ===
    vars = p, s, sn, d, dn, sp, sev, c = 'p', 's', 'sn', 'd', 'dn', 'sp', 'sev', 'c'

    symptom_query = inspect.cleandoc(f'''match
           $p isa person, has example-id {example_id};
           $s isa symptom, has name $sn;
           $d isa disease, has name $dn;
           $sp(presented-symptom: $s, symptomatic-patient: $p) isa symptom-presentation, has severity $sev;
           $c(cause: $d, effect: $s) isa causality;
           get;''')

    symptom_query_graph = (QueryGraph()
                           .add_vars(vars, PREEXISTS)
                           .add_has_edge(s, sn, PREEXISTS)
                           .add_has_edge(d, dn, PREEXISTS)
                           .add_role_edge(sp, s, 'presented-symptom', PREEXISTS)
                           .add_has_edge(sp, sev, PREEXISTS)
                           .add_role_edge(sp, p, 'symptomatic-patient', PREEXISTS)
                           .add_role_edge(c, s, 'effect', PREEXISTS)
                           .add_role_edge(c, d, 'cause', PREEXISTS))

    # === Diagnosis ===

    diag, d, p, dn = 'diag', 'd', 'p', 'dn'

    diagnosis_query = inspect.cleandoc(f'''match
           $p isa person, has example-id {example_id};
           $d isa disease, has name $dn;
           $diag(patient: $p, diagnosed-disease: $d) isa diagnosis;
           get;''')

    diagnosis_query_graph = (QueryGraph()
                             .add_vars([diag], TO_INFER)
                             .add_vars([d, p, dn], PREEXISTS)
                             .add_role_edge(diag, d, 'diagnosed-disease', TO_INFER)
                             .add_role_edge(diag, p, 'patient', TO_INFER))

    # === Candidate Diagnosis ===
    candidate_diagnosis_query = inspect.cleandoc(f'''match
           $p isa person, has example-id {example_id};
           $d isa disease, has name $dn;
           $diag(candidate-patient: $p, candidate-diagnosed-disease: $d) isa candidate-diagnosis; 
           get;''')

    candidate_diagnosis_query_graph = (QueryGraph()
                                       .add_vars([diag], CANDIDATE)
                                       .add_vars([d, p, dn], PREEXISTS)
                                       .add_role_edge(diag, d, 'candidate-diagnosed-disease', CANDIDATE)
                                       .add_role_edge(diag, p, 'candidate-patient', CANDIDATE))

    return [
        (symptom_query, lambda x: x, symptom_query_graph),
        (diagnosis_query, lambda x: x, diagnosis_query_graph),
        (candidate_diagnosis_query, lambda x: x, candidate_diagnosis_query_graph),
        (risk_factor_query, lambda x: x, risk_factor_query_graph),
        (person_age_query, lambda x: x, person_age_query_graph),
        (consumption_query, lambda x: x, consumption_query_graph),
        (hereditary_query, lambda x: x, hereditary_query_graph)
    ]


def write_predictions_to_grakn(graphs, tx):
    """
    Take predictions from the ML model, and insert representations of those predictions back into the graph.

    Args:
        graphs: graphs containing the concepts, with their class predictions and class probabilities
        tx: Grakn write transaction to use

    Returns: None

    """
    for graph in graphs:
        for node, data in graph.nodes(data=True):
            if data['prediction'] == 2:
                concept = data['concept']
                concept_type = concept.type_label
                if concept_type == 'diagnosis' or concept_type == 'candidate-diagnosis':
                    neighbours = graph.neighbors(node)

                    for neighbour in neighbours:
                        concept = graph.nodes[neighbour]['concept']
                        if concept.type_label == 'person':
                            person = concept
                        else:
                            disease = concept

                    p = data['probabilities']
                    query = (f'match'
                             f'$p id {person.id};'
                             f'$d id {disease.id};'
                             f'$kgcn isa kgcn;'
                             f'insert'
                             f'$pd(patient: $p, diagnosed-disease: $d, diagnoser: $kgcn) isa diagnosis,'
                             f'has probability-exists {p[2]:.3f},'
                             f'has probability-non-exists {p[1]:.3f},'  
                             f'has probability-preexists {p[0]:.3f};')
                    tx.query(query)
    tx.commit()


if __name__ == "__main__":
    diagnosis_example()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import six
from six.moves import queue


from grakn.service.Session.util.RequestBuilder import RequestBuilder
import grakn.service.Session.util.ResponseReader as ResponseReader # for circular import issue
from grakn.service.Session.util import enums
from grakn.exception.GraknError import GraknError

class TransactionService(object):

    def __init__(self, session_id, tx_type, transaction_endpoint):
        self.session_id = session_id
        self.tx_type = tx_type.value

        self._communicator = Communicator(transaction_endpoint)

        # open the transaction with an 'open' message
        open_req = RequestBuilder.open_tx(session_id, tx_type)
        self._communicator.send(open_req)
    __init__.__annotations__ = {'tx_type': enums.TxType}


    # --- Passthrough targets ---
    # targets of top level Transaction class

    def query(self, query, infer=True):
        request = RequestBuilder.query(query, infer=infer)
        # print("Query request: {0}".format(request))
        response = self._communicator.send(request)
        # convert `response` into a python iterator
        return ResponseReader.ResponseReader.query(self, response.query_iter)
    query.__annotations__ = {'query': str}

    def commit(self):
        request = RequestBuilder.commit()
        self._communicator.send(request)

    def close(self):
        self._communicator.close()

    def is_closed(self):
        return self._communicator._closed

    def get_concept(self, concept_id):
        request = RequestBuilder.get_concept(concept_id)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.get_concept(self, response.getConcept_res)
    get_concept.__annotations__ = {'concept_id': str}

    def get_schema_concept(self, label):
        request = RequestBuilder.get_schema_concept(label)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.get_schema_concept(self, response.getSchemaConcept_res)
    get_schema_concept.__annotations__ = {'label': str}

    def get_attributes_by_value(self, attribute_value, data_type):
        request = RequestBuilder.get_attributes_by_value(attribute_value, data_type)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.get_attributes_by_value(self, response.getAttributes_iter)
    get_attributes_by_value.__annotations__ = {'data_type': enums.DataType}

    def put_entity_type(self, label):
        request = RequestBuilder.put_entity_type(label)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.put_entity_type(self, response.putEntityType_res)
    put_entity_type.__annotations__ = {'label': str}

    def put_relation_type(self, label):
        request = RequestBuilder.put_relation_type(label)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.put_relation_type(self, response.putRelationType_res)
    put_relation_type.__annotations__ = {'label': str}

    def put_attribute_type(self, label, data_type):
        request = RequestBuilder.put_attribute_type(label, data_type)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.put_attribute_type(self, response.putAttributeType_res)
    put_attribute_type.__annotations__ = {'label': str, 'data_type': enums.DataType}

    def put_role(self, label):
        request = RequestBuilder.put_role(label)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.put_role(self, response.putRole_res)
    put_role.__annotations__ = {'label': str}

    def put_rule(self, label, when, then):
        request = RequestBuilder.put_rule(label, when, then)
        response = self._communicator.send(request)
        return ResponseReader.ResponseReader.put_rule(self, response.putRule_res)
    put_rule.__annotations__ = {'label': str, 'when': str, 'then': str}

    # --- Transaction Messages ---

    def run_concept_method(self, concept_id, grpc_concept_method_req):
        # wrap method_req into a transaction message
        tx_request = RequestBuilder.concept_method_req_to_tx_req(concept_id, grpc_concept_method_req)
        response = self._communicator.send(tx_request)
        return response.conceptMethod_res.response

    def explanation(self, explainable):
        """ Retrieve the explanation of a Concept Map from the server """
        tx_request = RequestBuilder.explanation(explainable)
        response = self._communicator.send(tx_request)
        return ResponseReader.ResponseReader.create_explanation(self, response.explanation_res)

    def iterate(self, iterator_id):
        request = RequestBuilder.next_iter(iterator_id)
        response = self._communicator.send(request)
        return response.iterate_res
    iterate.__annotations__ = {'iterator_id': int}


class Communicator(six.Iterator):
    """ An iterator and interface for GRPC stream """

    def __init__(self, grpc_stream_constructor):
        self._queue = queue.Queue()
        self._response_iterator = grpc_stream_constructor(self)
        self._closed = False

    def _add_request(self, request):
        self._queue.put(request)

    def __next__(self):
        # print("`next` called on Communicator")
        # print("Current queue: {0}".format(list(self._queue.queue)))
        next_item = self._queue.get(block=True)
        if next_item is None:
            raise StopIteration()
        return next_item

    def __iter__(self):
        return self

    def send(self, request):
        if self._closed:
            # TODO integrate this into TransactionService to throw a "Transaction is closed" rather than "connection is closed..."
            raise GraknError("This connection is closed")
        try:
            self._add_request(request)
            response = next(self._response_iterator)
        except Exception as e: # specialize into different gRPC exceptions?
            # invalidate this communicator, functionally this occurs automatically on exception (iterator not usable anymore)
            self._closed = True
            raise GraknError("Server/network error: {0}\n\n generated from request: {1}".format(e, request))

        if response is None:
            raise GraknError("No response received")
        
        return response

    def close(self):
        if not self._closed:
            with self._queue.mutex: # probably don't even need the mutex
                self._queue.queue.clear()
            self._queue.put(None)
            self._closed = True
            # force exhaust the iterator so `onCompleted()` is called on the server
            try:
                next(self._response_iterator)
            except StopIteration:
                pass



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import networkx as nx


def concept_dict_to_graph(concept_dict, variable_graph):
    """
    Create a new graph from a concept_dict, based on a `variable_graph` that describes the interactions between the
    variables in the `concept_dict`
    Args:
        concept_dict: A dictionary with variable names as keys, values are Things
        variable_graph: A graph with variable names as nodes, edges represent roles and has connections, indicated by
            the key "type" stored on each edge

    Returns:
        A graph with Things as nodes. Edges connecting the nodes have only role/has type information in their data
    """
    grakn_graph = nx.MultiDiGraph()
    node_to_var = {}

    if set(variable_graph.nodes()) != set(concept_dict.keys()):

        in_var_graph_not_in_concept_dict = set(variable_graph).difference(set(concept_dict.keys()))
        in_concept_dict_not_in_var_graph = set(concept_dict.keys()).difference(set(variable_graph))

        raise ValueError(f'The variables in the variable_graph must match those in the concept_dict\n'
                         f'In the variable graph but not in the concept dict: {in_var_graph_not_in_concept_dict}\n'
                         f'In the concept dict but not in the variable graph: {in_concept_dict_not_in_var_graph}')

    # This assumes that all variables are nodes, which would not be the case for variable roles
    for variable, thing in concept_dict.items():
        data = variable_graph.nodes[variable]
        data.update(type=thing.type_label)
        if thing.base_type_label == 'attribute':
            data.update(datatype=thing.data_type, value=thing.value)

        grakn_graph.add_node(thing, **data)

        # Record the mapping of nodes from one graph to the other
        assert variable not in node_to_var
        node_to_var[variable] = thing

    for sending_var, receiving_var, data in variable_graph.edges(data=True):
        sender = node_to_var[sending_var]
        receiver = node_to_var[receiving_var]

        if sender.base_type_label != 'relation' and not (
                receiver.base_type_label == 'attribute' and data['type'] == 'has'):
            raise ValueError('An edge in the variable_graph originates from a non-relation, check the variable_graph!')

        if data['type'] == 'has':
            grakn_graph.add_edge(sender, receiver, **data)
        else:
            grakn_graph.add_edge(sender, receiver, **data)

    return grakn_graph



import grpc

from grakn.service.Session.util.enums import DataType   # user-facing DataType enum

from grakn.service.Session.util.RequestBuilder import RequestBuilder
from grakn.service.Session.util.enums import TxType as _TxType
from grakn.service.Keyspace.KeyspaceService import KeyspaceService
from grakn.service.Session.TransactionService import TransactionService
from grakn_protocol.session.Session_pb2_grpc import SessionServiceStub
from grakn.exception.GraknError import GraknError

class GraknClient(object):
    """ A client/representation of a Grakn instance"""

    def __init__(self, uri, credentials=None):
        self.uri = uri
        self.credentials = credentials
        self._channel = grpc.insecure_channel(uri)
        self._keyspace_service = KeyspaceService(self.uri, self._channel, credentials)

    def session(self, keyspace):
        """ Open a session for a specific  keyspace. Can be used as `with Grakn('localhost:48555').session(keyspace='test') as session: ... ` or as normal assignment"""
        return Session(self.uri, keyspace, self._channel, self.credentials)
    session.__annotations__ = {'keyspace': str}

    def keyspaces(self):
        return self._keyspace_service

    def close(self):
        self._channel.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()
        if tb is None:
            # No exception
            pass
        else:
            #print("Closing Client due to exception: {0} \n traceback: \n {1}".format(type, tb))
            return False

class Session(object):
    """ A session for a Grakn instance and a specific keyspace """

    def __init__(self, uri, keyspace, channel, credentials):

        if not isinstance(uri, str):
            raise TypeError('expected string for uri')

        if not isinstance(keyspace, str):
            raise TypeError('expected string for keyspace')

        self.keyspace = keyspace
        self.uri = uri
        self.credentials = credentials

        self._stub = SessionServiceStub(channel)
        self._closed = False

        try:
            open_session_response = self._stub.open(RequestBuilder.open_session(keyspace, self.credentials))
            self.session_id = open_session_response.sessionId
        except Exception as e:
            raise GraknError('Could not obtain sessionId for keyspace "{0}", stems from: {1}'.format(keyspace, e))

    __init__.__annotations__ = {'uri': str, 'keyspace': str}

    def transaction(self):
        """ Build a read or write transaction to Grakn on this keyspace (ie. session.transaction().read() or .write()) """
        if self._closed:
            raise GraknError("Session is closed")

        # create a transaction service which hides GRPC usage
        return TransactionBuilder(self.session_id, self._stub.transaction)

    def close(self):
        """ Close this keyspace session """
        close_session_req = RequestBuilder.close_session(self.session_id)
        self._stub.close(close_session_req)
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()
        if tb is None:
            # No exception
            pass
        else:
            #print("Closing Session due to exception: {0} \n traceback: \n {1}".format(type, tb))
            return False

class TransactionBuilder(object):
    def __init__(self, session_id, transaction_rpc_constructor):
        self._session_id = session_id
        self._transaction_rpc_constructor = transaction_rpc_constructor

    def read(self):
        transaction_service = TransactionService(self._session_id, _TxType.READ, self._transaction_rpc_constructor)
        return Transaction(transaction_service)

    def write(self):
        transaction_service = TransactionService(self._session_id, _TxType.WRITE, self._transaction_rpc_constructor)
        return Transaction(transaction_service)


class Transaction(object):
    """ Presents the Grakn interface to the user, actual work with GRPC happens in TransactionService """

    def __init__(self, transaction_service):
        self._tx_service = transaction_service
    __init__.__annotations__ = {'transaction_service': TransactionService}

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()
        if tb is None:
            # No exception
            pass
        else:
            #print("Closing Transaction due to exception: {0} \n traceback: \n {1}".format(type, tb))
            return False

    def query(self, query, infer=True):
        """ Execute a Graql query, inference is optionally enabled """
        return self._tx_service.query(query, infer)
    query.__annotations__ = {'query': str}

    def commit(self):
        """ Commit and close this transaction, persisting changes to Grakn """
        self._tx_service.commit()
        self.close()

    def close(self):
        """ Close this transaction without committing """
        self._tx_service.close() # close the service

    def is_open(self):
        """ Check if this transaction is open"""
        return not self._tx_service.is_closed()

    def get_concept(self, concept_id):
        """ Retrieve a concept by Concept ID (string) """
        return self._tx_service.get_concept(concept_id)
    get_concept.__annotations__ = {'concept_id': str}

    def get_schema_concept(self, label):
        """ Retrieve a schema concept by its label (eg. those defined using `define` or tx.put...() """
        return self._tx_service.get_schema_concept(label)
    get_schema_concept.__annotations__ = {'label': str}

    def get_attributes_by_value(self, attribute_value, data_type):
        """ Retrieve atttributes with a specific value and datatype

        :param any attribute_value: the value to match
        :param grakn.DataType data_type: The data type of the value in Grakn, as given by the grakn.DataType enum
        """
        return self._tx_service.get_attributes_by_value(attribute_value, data_type)

    def put_entity_type(self, label):
        """ Define a new entity type with the given label """
        return self._tx_service.put_entity_type(label)
    put_entity_type.__annotations__ = {'label': str}

    def put_relation_type(self, label):
        """ Define a new relation type with the given label """
        return self._tx_service.put_relation_type(label)
    put_relation_type.__annotations__ = {'label': str}

    def put_attribute_type(self, label, data_type):
        """ Define a new attribute type with the given label and data type

        :param str label: the label of the attribute type
        :param grakn.DataType data_type: the data type of the value to be stored, as given by the grakn.DataType enum
        """
        return self._tx_service.put_attribute_type(label, data_type)
    put_attribute_type.__annotations__ = {'label': str}

    def put_role(self, label):
        """ Define a role with the given label """
        return self._tx_service.put_role(label)
    put_role.__annotations__ = {'label': str}

    def put_rule(self, label, when, then):
        """ Define a new rule with the given label, when and then clauses """
        return self._tx_service.put_rule(label, when, then)
    put_rule.__annotations__ = {'label': str, 'when': str, 'then': str}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from datetime import datetime

import grakn_protocol.session.Session_pb2 as transaction_messages
import grakn_protocol.session.Concept_pb2 as concept_messages
import grakn_protocol.session.Answer_pb2 as answer_messages
from grakn.service.Session.util import enums
from grakn.service.Session.Concept import BaseTypeMapping

import six

class RequestBuilder(object):
    """ Static methods for generating GRPC requests """

    @staticmethod
    def concept_method_req_to_tx_req(concept_id, grpc_concept_method_req):
        concept_method_req = transaction_messages.Transaction.ConceptMethod.Req()
        concept_method_req.id = concept_id
        concept_method_req.method.CopyFrom(grpc_concept_method_req)

        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.conceptMethod_req.CopyFrom(concept_method_req)
        return transaction_req

    # --- Top level functionality ---
    @staticmethod
    def open_tx(session_id, tx_type):
        open_request = transaction_messages.Transaction.Open.Req()
        open_request.sessionId = session_id
        open_request.type = tx_type.value

        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.open_req.CopyFrom(open_request)
        return transaction_req

    @staticmethod
    def open_session(keyspace, credentials):
        open_session_request = transaction_messages.Session.Open.Req()
        open_session_request.Keyspace = keyspace
        if credentials is not None:
            open_session_request.username = credentials['username']
            open_session_request.password = credentials['password']
        return open_session_request

    @staticmethod
    def close_session(session_id):
        close_session_request = transaction_messages.Session.Close.Req()
        close_session_request.sessionId = session_id
        return close_session_request

    @staticmethod
    def query(query, infer=True):
        query_message = transaction_messages.Transaction.Query.Req()
        query_message.query = query
        query_message.infer = transaction_messages.Transaction.Query.INFER.Value("TRUE") if infer else \
                              transaction_messages.Transaction.Query.INFER.Value("FALSE")
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.query_req.CopyFrom(query_message)
        return transaction_req

    @staticmethod
    def commit():
        commit_req = transaction_messages.Transaction.Commit.Req()
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.commit_req.CopyFrom(commit_req)
        return transaction_req


    @staticmethod
    def get_concept(concept_id):
        get_concept_req = transaction_messages.Transaction.GetConcept.Req()
        get_concept_req.id = concept_id
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.getConcept_req.CopyFrom(get_concept_req)
        return transaction_req

    @staticmethod
    def get_schema_concept(label):
        get_schema_concept_req = transaction_messages.Transaction.GetSchemaConcept.Req()
        get_schema_concept_req.label = label
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.getSchemaConcept_req.CopyFrom(get_schema_concept_req)
        return transaction_req

    def get_attributes_by_value(value, datatype):
        get_attrs_req = transaction_messages.Transaction.GetAttributes.Req()
        grpc_value_object = RequestBuilder.ConceptMethod.as_value_object(value, datatype)
        get_attrs_req.value.CopyFrom(grpc_value_object)
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.getAttributes_req.CopyFrom(get_attrs_req)
        return transaction_req
    get_attributes_by_value.__annotations__ = {'datatype': enums.DataType}
    get_attributes_by_value = staticmethod(get_attributes_by_value)


    @staticmethod
    def put_entity_type(label):
        put_entity_type_req = transaction_messages.Transaction.PutEntityType.Req()
        put_entity_type_req.label = label
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.putEntityType_req.CopyFrom(put_entity_type_req)
        return transaction_req

    @staticmethod
    def put_relation_type(label):
        put_relation_type_req = transaction_messages.Transaction.PutRelationType.Req()
        put_relation_type_req.label = label
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.putRelationType_req.CopyFrom(put_relation_type_req)
        return transaction_req

    def put_attribute_type(label, data_type):
        put_attribute_type_req = transaction_messages.Transaction.PutAttributeType.Req()
        put_attribute_type_req.label = label
        put_attribute_type_req.dataType = data_type.value # retrieve enum value
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.putAttributeType_req.CopyFrom(put_attribute_type_req)
        return transaction_req
    put_attribute_type.__annotations__ = {'data_type': enums.DataType}
    put_attribute_type = staticmethod(put_attribute_type)

    @staticmethod
    def put_role(label):
        put_role_req = transaction_messages.Transaction.PutRole.Req()
        put_role_req.label = label
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.putRole_req.CopyFrom(put_role_req)
        return transaction_req

    @staticmethod
    def put_rule(label, when, then):
        put_rule_req = transaction_messages.Transaction.PutRule.Req()
        put_rule_req.label = label
        put_rule_req.when = when
        put_rule_req.then = then
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.putRule_req.CopyFrom(put_rule_req)
        return transaction_req


    # --- internal requests ---

    @staticmethod
    def explanation(explainable):
        concept_map = {}
        for variable, concept in explainable.map().items():
            grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(concept)
            concept_map[variable] = grpc_concept

        grpc_concept_map = answer_messages.ConceptMap(map=concept_map)

        grpc_concept_map.hasExplanation = explainable.has_explanation()
        grpc_concept_map.pattern = explainable.query_pattern()

        explanation_req = answer_messages.Explanation.Req()
        explanation_req.explainable.CopyFrom(grpc_concept_map)

        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.explanation_req.CopyFrom(explanation_req)

        return transaction_req


    @staticmethod
    def next_iter(iterator_id):
        iterate_request = transaction_messages.Transaction.Iter.Req()
        iterate_request.id = iterator_id
        
        transaction_req = transaction_messages.Transaction.Req()
        transaction_req.iterate_req.CopyFrom(iterate_request)
        return transaction_req


    # ------ Concept Method Requests ------

    class ConceptMethod(object):
        """ Construct Concept Method requests """

        @staticmethod
        def delete():
            delete_req = concept_messages.Concept.Delete.Req()
            concept_method_req = concept_messages.Method.Req()
            concept_method_req.concept_delete_req.CopyFrom(delete_req)
            return concept_method_req

        @staticmethod
        def _concept_to_grpc_concept(concept):
            """ Takes a concept from ConceptHierarcy and converts to GRPC message """
            grpc_concept = concept_messages.Concept()
            grpc_concept.id = concept.id
            base_type_name = concept.base_type
            grpc_base_type = BaseTypeMapping.name_to_grpc_base_type[base_type_name]
            grpc_concept.baseType = grpc_base_type
            return grpc_concept

        def as_value_object(data, datatype):
            msg = concept_messages.ValueObject()
            if datatype == enums.DataType.STRING:
                msg.string = data
            elif datatype == enums.DataType.BOOLEAN:
                msg.boolean = data
            elif datatype == enums.DataType.INTEGER:
                msg.integer = data
            elif datatype == enums.DataType.LONG:
                msg.long = data
            elif datatype == enums.DataType.FLOAT:
                msg.float = data
            elif datatype == enums.DataType.DOUBLE:
                msg.double = data
            elif datatype == enums.DataType.DATE:
                # convert local datetime into long
                epoch = datetime(1970, 1, 1)
                diff = data - epoch
                epoch_seconds_utc = int(diff.total_seconds())
                epoch_ms_long_utc = int(epoch_seconds_utc*1000)
                msg.date = epoch_ms_long_utc 
            else:
                # TODO specialize exception
                raise Exception("Unknown attribute datatype: {}".format(datatype))
            return msg
        as_value_object.__annotations__ = {'datatype': enums.DataType}
        as_value_object = staticmethod(as_value_object)



        class SchemaConcept(object):
            """ Generates SchemaConcept method messages """

            @staticmethod
            def get_label():
                get_schema_label_req = concept_messages.SchemaConcept.GetLabel.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_getLabel_req.CopyFrom(get_schema_label_req)
                return concept_method_req
          
            @staticmethod
            def set_label(label):
                set_schema_label_req = concept_messages.SchemaConcept.SetLabel.Req()
                set_schema_label_req.label = label
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_setLabel_req.CopyFrom(set_schema_label_req)
                return concept_method_req

            @staticmethod
            def is_implicit():
                is_implicit_req = concept_messages.SchemaConcept.IsImplicit.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_isImplicit_req.CopyFrom(is_implicit_req)
                return concept_method_req

            @staticmethod
            def get_sup():
                get_sup_req = concept_messages.SchemaConcept.GetSup.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_getSup_req.CopyFrom(get_sup_req)
                return concept_method_req

            @staticmethod
            def set_sup(concept): 
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(concept)
                set_sup_req = concept_messages.SchemaConcept.SetSup.Req()
                set_sup_req.schemaConcept.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_setSup_req.CopyFrom(set_sup_req)
                return concept_method_req

            @staticmethod
            def subs():
                subs_req = concept_messages.SchemaConcept.Subs.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_subs_req.CopyFrom(subs_req)
                return concept_method_req

            @staticmethod
            def sups():
                sups_req = concept_messages.SchemaConcept.Sups.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.schemaConcept_sups_req.CopyFrom(sups_req)
                return concept_method_req



        class Rule(object):
            """ Generates Rule method messages """

            @staticmethod
            def when():
                when_req = concept_messages.Rule.When.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.rule_when_req.CopyFrom(when_req)
                return concept_method_req

            @staticmethod
            def then():
                then_req = concept_messages.Rule.Then.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.rule_then_req.CopyFrom(then_req)
                return concept_method_req

        class Role(object):
            """ Generates Role method messages """

            @staticmethod
            def relations():
                relations_req = concept_messages.Role.Relations.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.role_relations_req.CopyFrom(relations_req)
                return concept_method_req

            @staticmethod
            def players():
                players_req = concept_messages.Role.Players.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.role_players_req.CopyFrom(players_req)
                return concept_method_req

            

        class Type(object):
            """ Generates Type method messages """

            @staticmethod
            def is_abstract():
                is_abstract_req = concept_messages.Type.IsAbstract.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_isAbstract_req.CopyFrom(is_abstract_req)
                return concept_method_req

            @staticmethod
            def set_abstract(abstract):
                set_abstract_req = concept_messages.Type.SetAbstract.Req()
                assert type(abstract) == bool
                set_abstract_req.abstract = abstract
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_setAbstract_req.CopyFrom(set_abstract_req)
                return concept_method_req

            @staticmethod
            def instances():
                type_instances_req = concept_messages.Type.Instances.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_instances_req.CopyFrom(type_instances_req)
                return concept_method_req

            @staticmethod
            def keys():
                type_keys_req = concept_messages.Type.Keys.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_keys_req.CopyFrom(type_keys_req)
                return concept_method_req

            @staticmethod
            def attributes():
                type_attributes_req = concept_messages.Type.Attributes.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_attributes_req.CopyFrom(type_attributes_req)
                return concept_method_req 

            @staticmethod
            def has(attribute_type_concept):
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_type_concept)
                has_req = concept_messages.Type.Has.Req()
                has_req.attributeType.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_has_req.CopyFrom(has_req)
                return concept_method_req

            @staticmethod
            def unhas(attribute_type_concept):
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_type_concept)
                unhas_req = concept_messages.Type.Unhas.Req()
                unhas_req.attributeType.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_unhas_req.CopyFrom(unhas_req)
                return concept_method_req

            @staticmethod
            def key(attribute_type_concept):
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_type_concept)
                key_req = concept_messages.Type.Key.Req()
                key_req.attributeType.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_key_req.CopyFrom(key_req)
                return concept_method_req

            @staticmethod
            def unkey(attribute_type_concept):
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_type_concept)
                unkey_req = concept_messages.Type.Unkey.Req()
                unkey_req.attributeType.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_unkey_req.CopyFrom(unkey_req)
                return concept_method_req

            @staticmethod
            def playing():
                playing_req = concept_messages.Type.Playing.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_playing_req.CopyFrom(playing_req)
                return concept_method_req

            @staticmethod
            def plays( role_concept):
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                plays_req = concept_messages.Type.Plays.Req()
                plays_req.role.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_plays_req.CopyFrom(plays_req)
                return concept_method_req

            @staticmethod
            def unplay(role_concept):
                grpc_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                unplay_req = concept_messages.Type.Unplay.Req()
                unplay_req.role.CopyFrom(grpc_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.type_unplay_req.CopyFrom(unplay_req)
                return concept_method_req

       


        class EntityType(object):
            """ Generates EntityType method messages """

            @staticmethod
            def create():
                create_req = concept_messages.EntityType.Create.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.entityType_create_req.CopyFrom(create_req)
                return concept_method_req

        class RelationType(object):
            """ Generates RelationType method messages """
            
            @staticmethod
            def create():
                create_req = concept_messages.RelationType.Create.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.relationType_create_req.CopyFrom(create_req)
                return concept_method_req

            @staticmethod
            def roles():
                roles_req = concept_messages.RelationType.Roles.Req()
                concept_messages_req = concept_messages.Method.Req()
                concept_messages_req.relationType_roles_req.CopyFrom(roles_req)
                return concept_messages_req

            @staticmethod
            def relates(role_concept):
                grpc_role_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                relates_req = concept_messages.RelationType.Relates.Req()
                relates_req.role.CopyFrom(grpc_role_concept)
                concept_messages_req = concept_messages.Method.Req()
                concept_messages_req.relationType_relates_req.CopyFrom(relates_req)
                return concept_messages_req

            @staticmethod
            def unrelate(role_concept):
                grpc_role_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                unrelate_req = concept_messages.RelationType.Unrelate.Req()
                unrelate_req.role.CopyFrom(grpc_role_concept)
                concept_messages_req = concept_messages.Method.Req()
                concept_messages_req.relationType_unrelate_req.CopyFrom(unrelate_req)
                return concept_messages_req

        class AttributeType(object):
            """ Generates AttributeType method messages """
            
            @staticmethod
            def create(value, datatype):
                grpc_value_object = RequestBuilder.ConceptMethod.as_value_object(value, datatype)
                create_attr_req = concept_messages.AttributeType.Create.Req()
                create_attr_req.value.CopyFrom(grpc_value_object)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attributeType_create_req.CopyFrom(create_attr_req)
                return concept_method_req

            @staticmethod
            def attribute(value, datatype):
                grpc_value_object = RequestBuilder.ConceptMethod.as_value_object(value, datatype)
                attribute_req = concept_messages.AttributeType.Attribute.Req()
                attribute_req.value.CopyFrom(grpc_value_object)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attributeType_attribute_req.CopyFrom(attribute_req)
                return concept_method_req

            @staticmethod
            def data_type():
                datatype_req = concept_messages.AttributeType.DataType.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attributeType_dataType_req.CopyFrom(datatype_req)
                return concept_method_req

            @staticmethod
            def get_regex():
                get_regex_req = concept_messages.AttributeType.GetRegex.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attributeType_getRegex_req.CopyFrom(get_regex_req)
                return concept_method_req

            @staticmethod
            def set_regex(regex):
                set_regex_req = concept_messages.AttributeType.SetRegex.Req()
                set_regex_req.regex = regex
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attributeType_setRegex_req.CopyFrom(set_regex_req)
                return concept_method_req


        class Thing(object):
            """ Generates Thing method messages """

            @staticmethod
            def is_inferred():
                is_inferred_req = concept_messages.Thing.IsInferred.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_isInferred_req.CopyFrom(is_inferred_req)
                return concept_method_req

            @staticmethod
            def type():
                type_req = concept_messages.Thing.Type.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_type_req.CopyFrom(type_req)
                return concept_method_req
                
            @staticmethod
            def attributes(attribute_types=[]):
                """ Takes a list of AttributeType concepts to narrow attribute retrieval """
                attributes_req = concept_messages.Thing.Attributes.Req()
                for attribute_type_concept in attribute_types:
                    grpc_attr_type_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_type_concept)
                    attributes_req.attributeTypes.extend([grpc_attr_type_concept])
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_attributes_req.CopyFrom(attributes_req)
                return concept_method_req

            @staticmethod
            def relations(role_concepts=[]):
                """ Takes a list of role concepts to narrow the relations retrieval """
                relations_req = concept_messages.Thing.Relations.Req()
                for role_concept in role_concepts:
                    grpc_role_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                    # TODO this could use .add() if can be made to work...
                    relations_req.roles.extend([grpc_role_concept])
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_relations_req.CopyFrom(relations_req)
                return concept_method_req
            
            @staticmethod
            def roles():
                roles_req = concept_messages.Thing.Roles.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_roles_req.CopyFrom(roles_req)
                return concept_method_req

            @staticmethod
            def keys(attribute_types=[]):
                """ Takes a  list of AttributeType concepts to narrow the key retrieval """
                keys_req = concept_messages.Thing.Keys.Req()
                for attribute_type_concept in attribute_types:
                    grpc_attr_type_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_type_concept)
                    keys_req.attributeTypes.extend([grpc_attr_type_concept])
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_keys_req.CopyFrom(keys_req)
                return concept_method_req
            
            @staticmethod
            def has(attribute_concept):
                grpc_attribute_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_concept)
                relhas_req = concept_messages.Thing.Relhas.Req()
                relhas_req.attribute.CopyFrom(grpc_attribute_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_relhas_req.CopyFrom(relhas_req)
                return concept_method_req

            @staticmethod
            def unhas(attribute_concept):
                grpc_attribute_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(attribute_concept)
                unhas_req = concept_messages.Thing.Unhas.Req()
                unhas_req.attribute.CopyFrom(grpc_attribute_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.thing_unhas_req.CopyFrom(unhas_req)
                return concept_method_req
        
        class Relation(object):
            """ Generates Relation method messages """

            @staticmethod
            def role_players_map():
                role_players_map_req = concept_messages.Relation.RolePlayersMap.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.relation_rolePlayersMap_req.CopyFrom(role_players_map_req)
                return concept_method_req

            @staticmethod
            def role_players(roles=[]):
                """ Retrieve concepts that can play the given roles """
                role_players_req = concept_messages.Relation.RolePlayers.Req()
                for role_concept in roles:
                    grpc_role_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                    role_players_req.roles.extend([grpc_role_concept])
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.relation_rolePlayers_req.CopyFrom(role_players_req)
                return concept_method_req
    
            @staticmethod
            def assign(role_concept, player_concept):
                grpc_role_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                grpc_player_concept  = RequestBuilder.ConceptMethod._concept_to_grpc_concept(player_concept)
                assign_req = concept_messages.Relation.Assign.Req()
                assign_req.role.CopyFrom(grpc_role_concept)
                assign_req.player.CopyFrom(grpc_player_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.relation_assign_req.CopyFrom(assign_req)
                return concept_method_req
            
            @staticmethod
            def unassign(role_concept, player_concept):
                grpc_role_concept = RequestBuilder.ConceptMethod._concept_to_grpc_concept(role_concept)
                grpc_player_concept  = RequestBuilder.ConceptMethod._concept_to_grpc_concept(player_concept)
                unassign_req = concept_messages.Relation.Unassign.Req()
                unassign_req.role.CopyFrom(grpc_role_concept)
                unassign_req.player.CopyFrom(grpc_player_concept)
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.relation_unassign_req.CopyFrom(unassign_req)
                return concept_method_req

        class Attribute(object):
            """ Generates Attribute method messages """

            @staticmethod
            def value():
                value_req = concept_messages.Attribute.Value.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attribute_value_req.CopyFrom(value_req)
                return concept_method_req

            @staticmethod
            def owners():
                owners_req = concept_messages.Attribute.Owners.Req()
                concept_method_req = concept_messages.Method.Req()
                concept_method_req.attribute_owners_req.CopyFrom(owners_req)
                return concept_method_req
        
        class Entity(object):
            """ Empty implementation -- never create requests on Entity """
            pass



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import networkx.utils


def draw_networkx_labels(G, pos,
                         labels=None,
                         font_size=12,
                         font_color=None,
                         font_family='sans-serif',
                         font_weight='normal',
                         alpha=None,
                         bbox=None,
                         ax=None,
                         **kwds):
    """Draw node labels on the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.

    labels : dictionary, optional (default=None)
       Node labels in a dictionary keyed by node of text labels
       Node-keys in labels should appear as keys in `pos`.
       If needed use: `{n:lab for n,lab in labels.items() if n in pos}`

    font_size : int
       Font size for text labels (default=12)

    font_color : string
       Font color string (default='k' black)

    font_family : string
       Font family (default='sans-serif')

    font_weight : string
       Font weight (default='normal')

    alpha : float
       The text transparency (default=1.0)

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    Returns
    -------
    dict
        `dict` of labels keyed on the nodes

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> labels = nx.draw_networkx_labels(G, pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_edge_labels()
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib.cbook as cb
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()

    if labels is None:
        labels = dict((n, n) for n in G.nodes())

    # set optional alignment
    horizontalalignment = kwds.get('horizontalalignment', 'center')
    verticalalignment = kwds.get('verticalalignment', 'center')

    text_items = {}  # there is no text collection so we'll fake one
    for n, label in labels.items():
        (x, y) = pos[n]
        if not networkx.utils.is_string_like(label):
            label = str(label)  # this makes "1" and 1 labeled the same
        t = ax.text(x, y,
                    label,
                    size=font_size,
                    color=font_color[n],
                    family=font_family,
                    weight=font_weight,
                    alpha=alpha[n],
                    horizontalalignment=horizontalalignment,
                    verticalalignment=verticalalignment,
                    transform=ax.transData,
                    bbox=bbox,
                    clip_on=True,
                    )
        text_items[n] = t

    plt.tick_params(
        axis='both',
        which='both',
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False)

    return text_items


def draw_networkx_edge_labels(G, pos,
                              edge_labels=None,
                              label_pos=0.5,
                              font_size=10,
                              font_color=None,
                              font_family='sans-serif',
                              font_weight='normal',
                              alpha=None,
                              bbox=None,
                              ax=None,
                              rotate=True,
                              **kwds):
    """Draw edge labels.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    alpha : float
       The text transparency (default=1.0)

    edge_labels : dictionary
       Edge labels in a dictionary keyed by edge two-tuple of text
       labels (default=None). Only labels for the keys in the dictionary
       are drawn.

    label_pos : float
       Position of edge label along edge (0=head, 0.5=center, 1=tail)

    font_size : int
       Font size for text labels (default=12)

    font_color : string
       Font color string (default='k' black)

    font_weight : string
       Font weight (default='normal')

    font_family : string
       Font family (default='sans-serif')

    bbox : Matplotlib bbox
       Specify text box shape and colors.

    clip_on : bool
       Turn on clipping at axis boundaries (default=True)

    Returns
    -------
    dict
        `dict` of labels keyed on the edges

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> edge_labels = nx.draw_networkx_edge_labels(G, pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_labels()
    """
    try:
        import matplotlib.pyplot as plt
        import numpy as np
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()
    if edge_labels is None:
        labels = {(u, v): d for u, v, d in G.edges(data=True)}
    else:
        labels = edge_labels
    text_items = {}
    for (n1, n2), label in labels.items():
        (x1, y1) = pos[n1]
        (x2, y2) = pos[n2]
        (x, y) = (x1 * label_pos + x2 * (1.0 - label_pos),
                  y1 * label_pos + y2 * (1.0 - label_pos))

        if rotate:
            # in degrees
            angle = np.arctan2(y2 - y1, x2 - x1) / (2.0 * np.pi) * 360
            # make label orientation "right-side-up"
            if angle > 90:
                angle -= 180
            if angle < - 90:
                angle += 180
            # transform data coordinate angle to screen coordinate angle
            xy = np.array((x, y))
            trans_angle = ax.transData.transform_angles(np.array((angle,)),
                                                        xy.reshape((1, 2)))[0]
        else:
            trans_angle = 0.0
        # use default box of white with white border
        if bbox is None:
            bbox = dict(boxstyle='round',
                        ec=(1.0, 1.0, 1.0),
                        fc=(1.0, 1.0, 1.0),
                        )
        if not networkx.utils.is_string_like(label):
            label = str(label)  # this makes "1" and 1 labeled the same

        # set optional alignment
        horizontalalignment = kwds.get('horizontalalignment', 'center')
        verticalalignment = kwds.get('verticalalignment', 'center')

        t = ax.text(x, y,
                    label,
                    size=font_size,
                    color=font_color[(n1, n2)],
                    family=font_family,
                    weight=font_weight,
                    alpha=alpha[(n1, n2)],
                    horizontalalignment=horizontalalignment,
                    verticalalignment=verticalalignment,
                    rotation=trans_angle,
                    transform=ax.transData,
                    bbox=bbox,
                    zorder=1,
                    clip_on=True,
                    )
        text_items[(n1, n2)] = t

    plt.tick_params(
        axis='both',
        which='both',
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False)

    return text_items



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#


def duplicate_edges_in_reverse(graph):
    """
    Takes in a directed multi graph, and creates duplicates of all edges, the duplicates having reversed direction to
    the originals. This is useful since directed edges constrain the direction of messages passed. We want to permit
    omni-directional message passing.
    Args:
        graph: The graph

    Returns:
        The graph with duplicated edges, reversed, with all original edge properties attached to the duplicates
    """
    for sender, receiver, keys, data in graph.edges(data=True, keys=True):
        graph.add_edge(receiver, sender, keys, **data)
    return graph


def apply_logits_to_graphs(graph, logits_graph):
    """
    Take in a graph that describes the logits of the graph of interest, and store those logits on the graph as the
    property 'logits'. The graphs must correspond with one another

    Args:
        graph: Graph to apply logits to
        logits_graph: Graph containing logits

    Returns:
        graph with logits added as property 'logits'
    """

    for node, data in logits_graph.nodes(data=True):
        graph.nodes[node]['logits'] = list(data['features'])

    # TODO This is the desired implementation, but the graphs are altered by the model to have duplicated reversed
    #  edges, so this won't work for now
    # for sender, receiver, keys, data in logit_graph.edges(keys=True, data=True):
    #     graph.edges[sender, receiver, keys]['logits'] = list(data['features'])

    for sender, receiver, keys, data in graph.edges(keys=True, data=True):
        data['logits'] = list(logits_graph.edges[sender, receiver, keys]['features'])

    return graph



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import numpy as np
import sonnet as snt
from graph_nets import modules
from graph_nets import utils_tf
from graph_nets.modules import GraphIndependent


def softmax(x):
    return np.exp(x) / np.sum(np.exp(x))


def make_mlp_model(latent_size=16, num_layers=2):
    """Instantiates a new MLP, followed by LayerNorm.

    The parameters of each new MLP are not shared with others generated by
    this function.

    Returns:
      A Sonnet module which contains the MLP and LayerNorm.
    """
    return snt.Sequential([
        snt.nets.MLP([latent_size] * num_layers, activate_final=True),
        snt.LayerNorm()
    ])


class MLPGraphIndependent(snt.AbstractModule):
    """GraphIndependent with MLP edge, node, and global models."""

    def __init__(self, name="MLPGraphIndependent"):
        super(MLPGraphIndependent, self).__init__(name=name)
        with self._enter_variable_scope():
            self._network = GraphIndependent(
                edge_model_fn=make_mlp_model,
                node_model_fn=make_mlp_model)

    def _build(self, inputs):
        return self._network(inputs)


class MLPInteractionNetwork(snt.AbstractModule):
    """InteractionNetwork with MLP edge, node, and global models."""

    def __init__(self, name="MLPInteractionNetwork"):
        super(MLPInteractionNetwork, self).__init__(name=name)
        with self._enter_variable_scope():
            self._network = modules.InteractionNetwork(make_mlp_model, make_mlp_model)

    def _build(self, inputs):
        return self._network(inputs)


class KGCN(snt.AbstractModule):
    """
    A KGCN Neural Network with Message Passing. Implemented as a Sonnet Module.
    """

    def __init__(self,
                 thing_embedder,
                 role_embedder,
                 edge_output_size=3,
                 node_output_size=3,
                 latent_size=16,
                 num_layers=2,
                 name="KGCN"):
        super(KGCN, self).__init__(name=name)

        self._thing_embedder = thing_embedder
        self._role_embedder = role_embedder

        self._latent_size = latent_size
        self._num_layers = num_layers

        # Transforms the outputs into the appropriate shapes.
        if edge_output_size is None:
            edge_fn = None
        else:
            edge_fn = lambda: snt.Linear(edge_output_size, name="edge_output")
        if node_output_size is None:
            node_fn = None
        else:
            node_fn = lambda: snt.Linear(node_output_size, name="node_output")
        with self._enter_variable_scope():
            self._encoder = self._kg_encoder()
            self._core = MLPInteractionNetwork()
            self._decoder = MLPGraphIndependent()
            self._output_transform = modules.GraphIndependent(edge_fn, node_fn)

    def _edge_model(self):
        return snt.Sequential([self._role_embedder,
                               snt.nets.MLP([self._latent_size] * self._num_layers, activate_final=True),
                               snt.LayerNorm()])

    def _node_model(self):
        return snt.Sequential([self._thing_embedder,
                               snt.nets.MLP([self._latent_size] * self._num_layers, activate_final=True),
                               snt.LayerNorm()])

    def _kg_encoder(self):
        return GraphIndependent(self._edge_model, self._node_model, name='kg_encoder')

    def _build(self, input_op, num_processing_steps):
        latent = self._encoder(input_op)
        latent0 = latent
        output_ops = []
        for _ in range(num_processing_steps):
            core_input = utils_tf.concat([latent0, latent], axis=1)
            latent = self._core(core_input)
            decoded_op = self._decoder(latent)
            output_ops.append(self._output_transform(decoded_op))
        return output_ops



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from enum import Enum
import grakn_protocol.session.Session_pb2 as SessionMessages
import grakn_protocol.session.Concept_pb2 as ConceptMessages


class TxType(Enum):
    READ = SessionMessages.Transaction.Type.Value('READ')
    WRITE = SessionMessages.Transaction.Type.Value('WRITE')
    BATCH = SessionMessages.Transaction.Type.Value('BATCH')
    

data_type_map = {}

class DataType(Enum):
    STRING = ConceptMessages.AttributeType.DATA_TYPE.Value('STRING')
    BOOLEAN = ConceptMessages.AttributeType.DATA_TYPE.Value('BOOLEAN')
    INTEGER = ConceptMessages.AttributeType.DATA_TYPE.Value('INTEGER')
    LONG = ConceptMessages.AttributeType.DATA_TYPE.Value('LONG')
    FLOAT = ConceptMessages.AttributeType.DATA_TYPE.Value('FLOAT')
    DOUBLE = ConceptMessages.AttributeType.DATA_TYPE.Value('DOUBLE')
    DATE= ConceptMessages.AttributeType.DATA_TYPE.Value('DATE')



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import math

import graph_nets.utils_np as utils_np
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

import kglib.kgcn.plot.draw as custom_nx


def plot_across_training(logged_iterations, losses_tr, losses_ge, corrects_tr, corrects_ge, solveds_tr, solveds_ge,
                         output_file='./learning.png'):
    # Plot results curves.
    fig = plt.figure(1, figsize=(18, 3))
    fig.clf()
    x = np.array(logged_iterations)
    # Loss.
    y_tr = losses_tr
    y_ge = losses_ge
    ax = fig.add_subplot(1, 3, 1)
    ax.plot(x, y_tr, "k", label="Training")
    ax.plot(x, y_ge, "k--", label="Test/generalization")
    ax.set_title("Loss across training")
    ax.set_xlabel("Training iteration")
    ax.set_ylabel("Loss (binary cross-entropy)")
    ax.legend()
    # Correct.
    y_tr = corrects_tr
    y_ge = corrects_ge
    ax = fig.add_subplot(1, 3, 2)
    ax.plot(x, y_tr, "k", label="Training")
    ax.plot(x, y_ge, "k--", label="Test/generalization")
    ax.set_title("Fraction correct across training")
    ax.set_xlabel("Training iteration")
    ax.set_ylabel("Fraction nodes/edges correct")
    # Solved.
    y_tr = solveds_tr
    y_ge = solveds_ge
    ax = fig.add_subplot(1, 3, 3)
    ax.plot(x, y_tr, "k", label="Training")
    ax.plot(x, y_ge, "k--", label="Test/generalization")
    ax.set_title("Fraction solved across training")
    ax.set_xlabel("Training iteration")
    ax.set_ylabel("Fraction examples solved")

    plt.savefig(output_file, bbox_inches='tight')


def plot_predictions(raw_graphs, test_values, num_processing_steps_ge, solution_weights=(-0.5, 0.5, 0.5),
                     output_file='./graph.png'):

    # # Plot graphs and results after each processing step.
    # The white node is the start, and the black is the end. Other nodes are colored
    # from red to purple to blue, where red means the model is confident the node is
    # off the shortest path, blue means the model is confident the node is on the
    # shortest path, and purplish colors mean the model isn't sure.

    max_graphs_to_plot = 10
    num_steps_to_plot = 3
    node_size = 120

    num_graphs = len(raw_graphs)
    targets = utils_np.graphs_tuple_to_data_dicts(test_values["target"])
    step_indices = np.floor(
        np.linspace(0, num_processing_steps_ge - 1,
                    num_steps_to_plot)).astype(int).tolist()
    outputs = list(
        zip(*(utils_np.graphs_tuple_to_data_dicts(test_values["outputs"][i])
              for i in step_indices)))
    h = min(num_graphs, max_graphs_to_plot)
    w = num_steps_to_plot + 2
    fig = plt.figure(101, figsize=(18, h * 3))
    fig.clf()
    for j, (graph, target, output) in enumerate(zip(raw_graphs, targets, outputs)):
        if j >= h:
            break
        for s, r, d in graph.edges(data=True):
            d['weight'] = solution_weights[d['solution']]  # Looks good with high k
        pos = nx.spring_layout(graph, k=3 / math.sqrt(graph.number_of_nodes()), seed=1, weight='weight', iterations=50)
        # pos = nx.circular_layout(graph, scale=2)
        ground_truth_node_prob = target["nodes"][:, -1]
        ground_truth_edge_prob = target["edges"][:, -1]

        non_preexist_node_mask = mask_preexists(target["nodes"])
        non_preexist_edge_mask = mask_preexists(target["edges"])

        # Ground truth.
        iax = j * (2 + num_steps_to_plot) + 1
        ax = draw_subplot(graph, fig, pos, node_size, h, w, iax, ground_truth_node_prob, ground_truth_edge_prob, True)

        # Format the ground truth plot axes
        ax.set_axis_on()
        ax.set_xticks([])
        ax.set_yticks([])
        ax.spines['bottom'].set_color('blue')
        ax.spines['top'].set_color('blue')
        ax.spines['right'].set_color('blue')
        ax.spines['left'].set_color('blue')
        ax.grid(None)
        ax.set_title("Ground truth")

        # Prediction.
        for k, outp in enumerate(output):
            iax = j * (2 + num_steps_to_plot) + 2 + k
            node_prob = softmax_prob_last_dim(outp["nodes"]) * non_preexist_node_mask
            edge_prob = softmax_prob_last_dim(outp["edges"]) * non_preexist_edge_mask
            ax = draw_subplot(graph, fig, pos, node_size, h, w, iax, node_prob, edge_prob, False)
            ax.set_title("Model-predicted\nStep {:02d} / {:02d}".format(
                step_indices[k] + 1, step_indices[-1] + 1))

        # Class Winners
        # Displays whether the class represented by the last dimension was the winner
        node_prob = last_dim_was_class_winner(output[-1]["nodes"]) * non_preexist_node_mask
        edge_prob = last_dim_was_class_winner(output[-1]["edges"]) * non_preexist_edge_mask

        iax = j * (2 + num_steps_to_plot) + 2 + len(output)
        ax = draw_subplot(graph, fig, pos, node_size, h, w, iax, node_prob, edge_prob, False)

        # Format the class winners plot axes
        ax.set_axis_on()
        ax.set_xticks([])
        ax.set_yticks([])
        ax.spines['bottom'].set_color('green')
        ax.spines['top'].set_color('green')
        ax.spines['right'].set_color('green')
        ax.spines['left'].set_color('green')
        ax.grid(None)
        ax.set_title("Model-predicted winners")

    plt.savefig(output_file, bbox_inches='tight')


def mask_preexists(arr):
    return (arr[:, 0] == 0) * 1


def softmax_prob_last_dim(x):
    e = np.exp(x)
    return e[:, -1] / np.sum(e, axis=-1)


def last_dim_was_class_winner(x):
    return (np.argmax(x, axis=-1) == 2) * 1


def element_color(gt_plot, probability, element_props):
    """
    Determine the color values to use for a node/edge and its label
    gt plot:
    blue for existing elements, green for those to infer, red candidates

    output plot:
    blue for existing elements, green for those to infer, red for candidates, all with transparency
    """

    existing = 0
    candidate = 1
    to_infer = 2

    solution = element_props.get('solution')

    color_config = {
        to_infer: {'color': [0.0, 1.0, 0.0], 'gt_opacity': 1.0},
        candidate: {'color': [1.0, 0.0, 0.0], 'gt_opacity': 1.0},
        existing: {'color': [0.0, 0.0, 1.0], 'gt_opacity': 0.2}
    }

    chosen_config = color_config[solution]

    if gt_plot:
        opacity = chosen_config['gt_opacity']
    else:
        opacity = probability

    label = np.array([0.0, 0.0, 0.0] + [opacity])
    color = np.array(chosen_config['color'] + [opacity])

    return dict(element=color, label=label)


def draw_subplot(graph, fig, pos, node_size, h, w, iax, node_prob, edge_prob, gt_plot):
    ax = fig.add_subplot(h, w, iax)
    node_color = {}
    node_label_color = {}
    edge_color = {}
    edge_label_color = {}

    for i, (n, props) in enumerate(graph.nodes(data=True)):
        colors = element_color(gt_plot, node_prob[n], props)

        node_color[n] = colors['element']
        node_label_color[n] = colors['label']

    for n, (sender, receiver, props) in enumerate(graph.edges(data=True)):
        colors = element_color(gt_plot, edge_prob[n], props)

        edge_color[(sender, receiver)] = colors['element']
        edge_label_color[(sender, receiver)] = colors['label']

    draw_graph(graph, pos, ax, node_size=node_size, node_color=node_color, node_label_color=node_label_color,
               edge_color=edge_color, edge_label_color=edge_label_color)
    return ax


def draw_graph(graph,
               pos,
               ax,
               node_size=200,
               node_color=(0.4, 0.8, 0.4),
               node_label_color=None,
               edge_color=(0.0, 0.0, 0.0),
               edge_label_color=None,
               node_linewidth=1.0,
               edge_width=1.0,
               font_size=6):

    def _draw(draw_function, zorder=None, **kwargs):
        # draw_kwargs = self._make_draw_kwargs(**kwargs)
        _base_draw_kwargs = dict(G=graph, pos=pos, ax=ax)
        kwargs.update(_base_draw_kwargs)
        collection = draw_function(**kwargs)
        if collection is not None and zorder is not None:
            try:
                # This is for compatibility with older matplotlib.
                collection.set_zorder(zorder)
            except AttributeError:
                # This is for compatibility with newer matplotlib.
                collection[0].set_zorder(zorder)
        return collection

    # Plot nodes.
    c = [node_color[n] for n in graph.nodes()]
    _draw(nx.draw_networkx_nodes,
          node_size=node_size,
          node_color=c,
          linewidths=node_linewidth,
          alpha=[node_color[n][-1] for n in graph.nodes()],
          zorder=-10)

    # Plot edges.
    e = [edge_color[(s, r)] for s, r, k in graph.edges]
    _draw(nx.draw_networkx_edges,
          edgelist=graph.edges,
          width=edge_width,
          zorder=-20,
          edge_color=e
          )

    bbox_props = dict(boxstyle="square,pad=0.0", fc="none", ec="none", lw=1)
    labels_dict = {node_id: graph.nodes[node_id]['type'] for node_id in graph.nodes}
    edge_labels_dict = {(edge_id[0], edge_id[1]): graph.edges[edge_id]['type'] for edge_id in graph.edges}

    custom_nx.draw_networkx_labels(graph,
                                   pos,
                                   labels=labels_dict,
                                   font_size=font_size,
                                   font_color=node_label_color,
                                   alpha=[node_label_color[n][-1] for n in graph.nodes()])

    custom_nx.draw_networkx_edge_labels(graph,
                                        pos,
                                        edge_labels=edge_labels_dict,
                                        font_size=font_size,
                                        # font_color=np.array([0.0, 0.5, 0.0, 0.1]),
                                        font_color=edge_label_color,
                                        # alpha=0.2,
                                        alpha={n: edge_label_color[n][-1] for n in graph.edges()},
                                        bbox=bbox_props)



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import time

import tensorflow as tf

from kglib.kgcn.learn.feed import create_placeholders, create_feed_dict, make_all_runnable_in_session
from kglib.kgcn.learn.loss import loss_ops_preexisting_no_penalty
from kglib.kgcn.learn.metrics import existence_accuracy


class KGCNLearner:
    """
    Responsible for running a KGCN model
    """
    def __init__(self, model, num_processing_steps_tr=10, num_processing_steps_ge=10):
        self._model = model
        self._num_processing_steps_tr = num_processing_steps_tr
        self._num_processing_steps_ge = num_processing_steps_ge

    def __call__(self,
                 tr_input_graphs,
                 tr_target_graphs,
                 ge_input_graphs,
                 ge_target_graphs,
                 num_training_iterations=1000,
                 learning_rate=1e-3,
                 log_every_epochs=20,
                 log_dir=None):
        """
        Args:
            tr_graphs: In-memory graphs of Grakn concepts for training
            ge_graphs: In-memory graphs of Grakn concepts for generalisation
            num_processing_steps_tr: Number of processing (message-passing) steps for training.
            num_processing_steps_ge: Number of processing (message-passing) steps for generalization.
            num_training_iterations: Number of training iterations
            log_every_seconds: The time to wait between logging and printing the next set of results.
            log_dir: Directory to store TensorFlow events files

        Returns:

        """

        tf.set_random_seed(1)

        input_ph, target_ph = create_placeholders(tr_input_graphs, tr_target_graphs)

        # A list of outputs, one per processing step.
        output_ops_tr = self._model(input_ph, self._num_processing_steps_tr)
        output_ops_ge = self._model(input_ph, self._num_processing_steps_ge)

        # Training loss.
        loss_ops_tr = loss_ops_preexisting_no_penalty(target_ph, output_ops_tr)
        # Loss across processing steps.
        loss_op_tr = sum(loss_ops_tr) / self._num_processing_steps_tr

        tf.summary.scalar('loss_op_tr', loss_op_tr)
        # Test/generalization loss.
        loss_ops_ge = loss_ops_preexisting_no_penalty(target_ph, output_ops_ge)
        loss_op_ge = loss_ops_ge[-1]  # Loss from final processing step.
        tf.summary.scalar('loss_op_ge', loss_op_ge)

        # Optimizer
        optimizer = tf.train.AdamOptimizer(learning_rate)
        gradients, variables = zip(*optimizer.compute_gradients(loss_op_tr))

        for grad, var in zip(gradients, variables):
            try:
                print(var.name)
                tf.summary.histogram('gradients/' + var.name, grad)
            except:
                pass

        gradients, _ = tf.clip_by_global_norm(gradients, 5.0)
        step_op = optimizer.apply_gradients(zip(gradients, variables))

        input_ph, target_ph = make_all_runnable_in_session(input_ph, target_ph)

        sess = tf.Session()
        merged_summaries = tf.summary.merge_all()

        train_writer = None

        if log_dir is not None:
            train_writer = tf.summary.FileWriter(log_dir, sess.graph)

        sess.run(tf.global_variables_initializer())

        logged_iterations = []
        losses_tr = []
        corrects_tr = []
        solveds_tr = []
        losses_ge = []
        corrects_ge = []
        solveds_ge = []

        print("# (iteration number), T (elapsed seconds), "
              "Ltr (training loss), Lge (test/generalization loss), "
              "Ctr (training fraction nodes/edges labeled correctly), "
              "Str (training fraction examples solved correctly), "
              "Cge (test/generalization fraction nodes/edges labeled correctly), "
              "Sge (test/generalization fraction examples solved correctly)")

        start_time = time.time()
        for iteration in range(num_training_iterations):
            feed_dict = create_feed_dict(input_ph, target_ph, tr_input_graphs, tr_target_graphs)

            if iteration % log_every_epochs == 0:

                train_values = sess.run(
                    {
                        "step": step_op,
                        "target": target_ph,
                        "loss": loss_op_tr,
                        "outputs": output_ops_tr,
                        "summary": merged_summaries
                    },
                    feed_dict=feed_dict)

                if train_writer is not None:
                    train_writer.add_summary(train_values["summary"], iteration)

                feed_dict = create_feed_dict(input_ph, target_ph, ge_input_graphs, ge_target_graphs)
                test_values = sess.run(
                    {
                        "target": target_ph,
                        "loss": loss_op_ge,
                        "outputs": output_ops_ge
                    },
                    feed_dict=feed_dict)
                correct_tr, solved_tr = existence_accuracy(
                    train_values["target"], train_values["outputs"][-1], use_edges=False)
                correct_ge, solved_ge = existence_accuracy(
                    test_values["target"], test_values["outputs"][-1], use_edges=False)

                elapsed = time.time() - start_time
                losses_tr.append(train_values["loss"])
                corrects_tr.append(correct_tr)
                solveds_tr.append(solved_tr)
                losses_ge.append(test_values["loss"])
                corrects_ge.append(correct_ge)
                solveds_ge.append(solved_ge)
                logged_iterations.append(iteration)
                print("# {:05d}, T {:.1f}, Ltr {:.4f}, Lge {:.4f}, Ctr {:.4f}, Str"
                      " {:.4f}, Cge {:.4f}, Sge {:.4f}".format(
                        iteration, elapsed, train_values["loss"], test_values["loss"],
                        correct_tr, solved_tr, correct_ge, solved_ge))
            else:
                train_values = sess.run(
                    {
                        "step": step_op,
                        "target": target_ph,
                        "loss": loss_op_tr,
                        "outputs": output_ops_tr
                    },
                    feed_dict=feed_dict)

        training_info = logged_iterations, losses_tr, losses_ge, corrects_tr, corrects_ge, solveds_tr, solveds_ge
        return train_values, test_values, training_info



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keyspace/Keyspace.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='keyspace/Keyspace.proto',
  package='keyspace',
  syntax='proto3',
  serialized_options=_b('\n\027grakn.protocol.keyspaceB\rKeyspaceProto'),
  serialized_pb=_b('\n\x17keyspace/Keyspace.proto\x12\x08keyspace\"\xd6\x01\n\x08Keyspace\x1a\x66\n\x08Retrieve\x1a=\n\x03Req\x12\x1a\n\x08username\x18\x01 \x01(\tR\x08username\x12\x1a\n\x08password\x18\x02 \x01(\tR\x08password\x1a\x1b\n\x03Res\x12\x14\n\x05names\x18\x01 \x03(\tR\x05names\x1a\x62\n\x06\x44\x65lete\x1aQ\n\x03Req\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1a\n\x08username\x18\x02 \x01(\tR\x08username\x12\x1a\n\x08password\x18\x03 \x01(\tR\x08password\x1a\x05\n\x03Res2\xa7\x01\n\x0fKeyspaceService\x12L\n\x08retrieve\x12\x1f.keyspace.Keyspace.Retrieve.Req\x1a\x1f.keyspace.Keyspace.Retrieve.Res\x12\x46\n\x06\x64\x65lete\x12\x1d.keyspace.Keyspace.Delete.Req\x1a\x1d.keyspace.Keyspace.Delete.ResB(\n\x17grakn.protocol.keyspaceB\rKeyspaceProtob\x06proto3')
)




_KEYSPACE_RETRIEVE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='keyspace.Keyspace.Retrieve.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='username', full_name='keyspace.Keyspace.Retrieve.Req.username', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='username', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='password', full_name='keyspace.Keyspace.Retrieve.Req.password', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='password', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=123,
)

_KEYSPACE_RETRIEVE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='keyspace.Keyspace.Retrieve.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='keyspace.Keyspace.Retrieve.Res.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='names', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=152,
)

_KEYSPACE_RETRIEVE = _descriptor.Descriptor(
  name='Retrieve',
  full_name='keyspace.Keyspace.Retrieve',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_KEYSPACE_RETRIEVE_REQ, _KEYSPACE_RETRIEVE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=50,
  serialized_end=152,
)

_KEYSPACE_DELETE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='keyspace.Keyspace.Delete.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='keyspace.Keyspace.Delete.Req.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='username', full_name='keyspace.Keyspace.Delete.Req.username', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='username', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='password', full_name='keyspace.Keyspace.Delete.Req.password', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='password', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=164,
  serialized_end=245,
)

_KEYSPACE_DELETE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='keyspace.Keyspace.Delete.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=130,
)

_KEYSPACE_DELETE = _descriptor.Descriptor(
  name='Delete',
  full_name='keyspace.Keyspace.Delete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_KEYSPACE_DELETE_REQ, _KEYSPACE_DELETE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=154,
  serialized_end=252,
)

_KEYSPACE = _descriptor.Descriptor(
  name='Keyspace',
  full_name='keyspace.Keyspace',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_KEYSPACE_RETRIEVE, _KEYSPACE_DELETE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=252,
)

_KEYSPACE_RETRIEVE_REQ.containing_type = _KEYSPACE_RETRIEVE
_KEYSPACE_RETRIEVE_RES.containing_type = _KEYSPACE_RETRIEVE
_KEYSPACE_RETRIEVE.containing_type = _KEYSPACE
_KEYSPACE_DELETE_REQ.containing_type = _KEYSPACE_DELETE
_KEYSPACE_DELETE_RES.containing_type = _KEYSPACE_DELETE
_KEYSPACE_DELETE.containing_type = _KEYSPACE
DESCRIPTOR.message_types_by_name['Keyspace'] = _KEYSPACE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Keyspace = _reflection.GeneratedProtocolMessageType('Keyspace', (_message.Message,), dict(

  Retrieve = _reflection.GeneratedProtocolMessageType('Retrieve', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _KEYSPACE_RETRIEVE_REQ,
      __module__ = 'keyspace.Keyspace_pb2'
      # @@protoc_insertion_point(class_scope:keyspace.Keyspace.Retrieve.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _KEYSPACE_RETRIEVE_RES,
      __module__ = 'keyspace.Keyspace_pb2'
      # @@protoc_insertion_point(class_scope:keyspace.Keyspace.Retrieve.Res)
      ))
    ,
    DESCRIPTOR = _KEYSPACE_RETRIEVE,
    __module__ = 'keyspace.Keyspace_pb2'
    # @@protoc_insertion_point(class_scope:keyspace.Keyspace.Retrieve)
    ))
  ,

  Delete = _reflection.GeneratedProtocolMessageType('Delete', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _KEYSPACE_DELETE_REQ,
      __module__ = 'keyspace.Keyspace_pb2'
      # @@protoc_insertion_point(class_scope:keyspace.Keyspace.Delete.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _KEYSPACE_DELETE_RES,
      __module__ = 'keyspace.Keyspace_pb2'
      # @@protoc_insertion_point(class_scope:keyspace.Keyspace.Delete.Res)
      ))
    ,
    DESCRIPTOR = _KEYSPACE_DELETE,
    __module__ = 'keyspace.Keyspace_pb2'
    # @@protoc_insertion_point(class_scope:keyspace.Keyspace.Delete)
    ))
  ,
  DESCRIPTOR = _KEYSPACE,
  __module__ = 'keyspace.Keyspace_pb2'
  # @@protoc_insertion_point(class_scope:keyspace.Keyspace)
  ))
_sym_db.RegisterMessage(Keyspace)
_sym_db.RegisterMessage(Keyspace.Retrieve)
_sym_db.RegisterMessage(Keyspace.Retrieve.Req)
_sym_db.RegisterMessage(Keyspace.Retrieve.Res)
_sym_db.RegisterMessage(Keyspace.Delete)
_sym_db.RegisterMessage(Keyspace.Delete.Req)
_sym_db.RegisterMessage(Keyspace.Delete.Res)


DESCRIPTOR._options = None

_KEYSPACESERVICE = _descriptor.ServiceDescriptor(
  name='KeyspaceService',
  full_name='keyspace.KeyspaceService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=255,
  serialized_end=422,
  methods=[
  _descriptor.MethodDescriptor(
    name='retrieve',
    full_name='keyspace.KeyspaceService.retrieve',
    index=0,
    containing_service=None,
    input_type=_KEYSPACE_RETRIEVE_REQ,
    output_type=_KEYSPACE_RETRIEVE_RES,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='delete',
    full_name='keyspace.KeyspaceService.delete',
    index=1,
    containing_service=None,
    input_type=_KEYSPACE_DELETE_REQ,
    output_type=_KEYSPACE_DELETE_RES,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_KEYSPACESERVICE)

DESCRIPTOR.services_by_name['KeyspaceService'] = _KEYSPACESERVICE

# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#


from grakn.service.Session.util import enums
from grakn.service.Session.util.RequestBuilder import RequestBuilder
from grakn.exception.GraknError import GraknError 



class Concept(object):

    def __init__(self, concept_id, base_type, tx_service):
        self.id = concept_id
        self.base_type = base_type
        self._tx_service = tx_service


    def delete(self):
        del_request = RequestBuilder.ConceptMethod.delete()
        method_response = self._tx_service.run_concept_method(self.id, del_request)
        return

    def is_deleted(self):
        retrieved = self._tx_service.get_concept(self.id)
        return retrieved is None


    def is_schema_concept(self):
        """ Check if this concept is a schema concept """
        return isinstance(self, SchemaConcept)
    is_schema_concept.__annotations__ = {'return': bool}

    def is_type(self):
        """ Check if this concept is a Type concept """
        return isinstance(self, Type)
    is_type.__annotations__ = {'return': bool}

    def is_thing(self):
        """ Check if this concept is a Thing concept """
        return isinstance(self, Thing)
    is_thing.__annotations__ = {'return': bool}

    def is_attribute_type(self):
        """ Check if this concept is an AttributeType concept """
        return isinstance(self, AttributeType)
    is_attribute_type.__annotations__ = {'return': bool}

    def is_entity_type(self):
        """ Check if this concept is an EntityType concept """
        return isinstance(self, EntityType)
    is_entity_type.__annotations__ = {'return': bool}

    def is_relation_type(self):
        """ Check if this concept is a RelationType concept """
        return isinstance(self, RelationType)
    is_relation_type.__annotations__ = {'return': bool}

    def is_role(self):
        """ Check if this concept is a Role """
        return isinstance(self, Role)
    is_role.__annotations__ = {'return': bool}

    def is_rule(self):
        """ Check if this concept is a Rule concept """
        return isinstance(self, Rule)
    is_rule.__annotations__ = {'return': bool}

    def is_attribute(self):
        """ Check if this concept is an Attribute concept """
        return isinstance(self, Attribute)
    is_attribute.__annotations__ = {'return': bool}

    def is_entity(self):
        """ Check if this concept is an Entity concept """
        return isinstance(self, Entity)
    is_entity.__annotations__ = {'return': bool}

    def is_relation(self):
        """ Check if this concept is a Relation concept """
        return isinstance(self, Relation)
    is_relation.__annotations__ = {'return': bool}


class SchemaConcept(Concept):

    def label(self, value=None):
        """ 
        Get or set label of this schema concept. 
        If used as setter returns self 
        """
        if value is None:
            get_label_req = RequestBuilder.ConceptMethod.SchemaConcept.get_label()
            method_response = self._tx_service.run_concept_method(self.id, get_label_req)
            return method_response.schemaConcept_getLabel_res.label
        else:
            set_label_req = RequestBuilder.ConceptMethod.SchemaConcept.set_label(value)
            method_response = self._tx_service.run_concept_method(self.id, set_label_req)
            return self

    def is_implicit(self):
        """ Check if this schema concept is implicit """
        is_implicit_req = RequestBuilder.ConceptMethod.SchemaConcept.is_implicit()
        method_response = self._tx_service.run_concept_method(self.id, is_implicit_req)
        return method_response.schemaConcept_isImplicit_res.implicit

    def sup(self, super_concept=None):
        """ 
        Get or set super schema concept.
        If used as a setter returns self
        """
        if super_concept is None:
            # get direct super schema concept
            get_sup_req = RequestBuilder.ConceptMethod.SchemaConcept.get_sup()
            method_response = self._tx_service.run_concept_method(self.id, get_sup_req)
            get_sup_response = method_response.schemaConcept_getSup_res 
            # check if received a Null or Concept
            whichone = get_sup_response.WhichOneof('res')
            if whichone == 'schemaConcept':
                grpc_schema_concept = get_sup_response.schemaConcept
                from grakn.service.Session.Concept import ConceptFactory
                concept = ConceptFactory.create_concept(self._tx_service, grpc_schema_concept)
                return concept
            elif whichone == 'null':
                return None
            else:
                raise GraknError("Unknown response concent for getting super schema concept: {0}".format(whichone))
        else:
            # set direct super SchemaConcept of this SchemaConcept
            set_sup_req = RequestBuilder.ConceptMethod.SchemaConcept.set_sup(super_concept)
            method_response = self._tx_service.run_concept_method(self.id, set_sup_req)
            return self

    def subs(self):
        """ Retrieve the sub schema concepts of this schema concept, as an iterator """
        subs_req = RequestBuilder.ConceptMethod.SchemaConcept.subs()
        method_response = self._tx_service.run_concept_method(self.id, subs_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                    self._tx_service,
                    method_response.schemaConcept_subs_iter.id,
                    lambda tx_serv, iter_res: 
                        ConceptFactory.create_concept(tx_serv,
                        iter_res.conceptMethod_iter_res.schemaConcept_subs_iter_res.schemaConcept)
                    )

    def sups(self):
        """ Retrieve the all supertypes (direct and higher level) of this schema concept as an iterator """
        sups_req = RequestBuilder.ConceptMethod.SchemaConcept.sups()
        method_response = self._tx_service.run_concept_method(self.id, sups_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.schemaConcept_sups_iter.id,
                lambda tx_serv, iter_res:
                    ConceptFactory.create_concept(tx_serv,
                    iter_res.conceptMethod_iter_res.schemaConcept_sups_iter_res.schemaConcept)
                )
 
class Type(SchemaConcept):

    def is_abstract(self, value=None):
        """
        Get/Set whether this schema Type object is abstract.
        When used as a setter returns `self` 
        """
        if value is None:
            # return True/False if the type is set to abstract
            is_abstract_req = RequestBuilder.ConceptMethod.Type.is_abstract()
            method_response = self._tx_service.run_concept_method(self.id, is_abstract_req)
            return method_response.type_isAbstract_res.abstract
        else:
            set_abstract_req = RequestBuilder.ConceptMethod.Type.set_abstract(value)
            method_response = self._tx_service.run_concept_method(self.id, set_abstract_req)
            return self
    is_abstract.__annotations__ = {'value': bool, 'return': bool}

    def attributes(self):
        """ Retrieve all attributes attached to this Type as an iterator """
        attributes_req = RequestBuilder.ConceptMethod.Type.attributes()
        method_response = self._tx_service.run_concept_method(self.id, attributes_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.type_attributes_iter.id,
                lambda tx_serv, iter_res: 
                    ConceptFactory.create_concept(tx_serv,
                    iter_res.conceptMethod_iter_res.type_attributes_iter_res.attributeType)
                )

    def instances(self):
        """ Retrieve all instances of this Type as an iterator """
        instances_req = RequestBuilder.ConceptMethod.Type.instances()
        method_response = self._tx_service.run_concept_method(self.id, instances_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.type_instances_iter.id,
                lambda tx_serv, iter_res: 
                    ConceptFactory.create_concept(tx_serv,
                    iter_res.conceptMethod_iter_res.type_instances_iter_res.thing)
                )

    def playing(self):
        """ Retrieve iterator of roles played by this type """
        playing_req = RequestBuilder.ConceptMethod.Type.playing()
        method_response = self._tx_service.run_concept_method(self.id, playing_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.type_playing_iter.id,
                lambda tx_serv, iter_res:
                    ConceptFactory.create_concept(tx_serv,
                    iter_res.conceptMethod_iter_res.type_playing_iter_res.role)
                )

    def plays(self, role_concept):
        """ Set a role that is played by this Type """
        plays_req = RequestBuilder.ConceptMethod.Type.plays(role_concept)
        method_response = self._tx_service.run_concept_method(self.id, plays_req)
        return self

    def unplay(self, role_concept):
        """ Remove a role that is played by this Type """
        unplay_req = RequestBuilder.ConceptMethod.Type.unplay(role_concept)
        method_response = self._tx_service.run_concept_method(self.id, unplay_req)
        return
    
    def has(self, attribute_concept_type):
        """ Attach an attributeType concept to the type """
        has_req = RequestBuilder.ConceptMethod.Type.has(attribute_concept_type)
        method_response = self._tx_service.run_concept_method(self.id, has_req)
        return self
        
    def unhas(self, attribute_concept_type):
        """ Remove an attribute type concept from this type """
        unhas_req = RequestBuilder.ConceptMethod.Type.unhas(attribute_concept_type)
        method_response = self._tx_service.run_concept_method(self.id, unhas_req)
        return self

    def keys(self):
        """ Retrieve an iterator of attribute types that this Type uses as keys """
        keys_req = RequestBuilder.ConceptMethod.Type.keys()
        method_response = self._tx_service.run_concept_method(self.id, keys_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.type_keys_iter.id,
                lambda tx_serv, iter_res:
                    ConceptFactory.create_concept(tx_serv,
                    iter_res.conceptMethod_iter_res.type_keys_iter_res.attributeType)
                )


    def key(self, attribute_concept_type):
        """ Add an attribute type to be a key for this Type """
        key_req = RequestBuilder.ConceptMethod.Type.key(attribute_concept_type)
        method_response = self._tx_service.run_concept_method(self.id, key_req)
        return self

    def unkey(self, attribute_concept_type):
        """ Remove an attribute type from this Type from being a key """
        unkey_req = RequestBuilder.ConceptMethod.Type.unkey(attribute_concept_type)
        method_response = self._tx_service.run_concept_method(self.id, unkey_req)
        return self



class EntityType(Type):

    def create(self):
        """ Instantiate an entity of the given type and return it """
        create_req = RequestBuilder.ConceptMethod.EntityType.create()
        method_response = self._tx_service.run_concept_method(self.id, create_req)
        grpc_entity_concept = method_response.entityType_create_res.entity
        from grakn.service.Session.Concept import ConceptFactory
        return ConceptFactory.create_concept(self._tx_service, grpc_entity_concept)

class AttributeType(Type):
    
    def create(self, value):
        """ Create an instance with this AttributeType """
        self_data_type = self.data_type()
        create_inst_req = RequestBuilder.ConceptMethod.AttributeType.create(value, self_data_type)
        method_response = self._tx_service.run_concept_method(self.id, create_inst_req)
        grpc_attribute_concept = method_response.attributeType_create_res.attribute
        from grakn.service.Session.Concept import ConceptFactory
        return ConceptFactory.create_concept(self._tx_service, grpc_attribute_concept)
        
    def attribute(self, value):
        """ Retrieve an attribute instance by value if it exists """
        self_data_type = self.data_type()
        get_attribute_req = RequestBuilder.ConceptMethod.AttributeType.attribute(value, self_data_type)
        method_response = self._tx_service.run_concept_method(self.id, get_attribute_req)
        response = method_response.attributeType_attribute_res
        whichone = response.WhichOneof('res')
        if whichone == 'attribute':
            from grakn.service.Session.Concept import ConceptFactory
            return ConceptFactory.create_concept(self._tx_service, response.attribute)
        elif whichone == 'null':
            return None
        else:
            raise GraknError("Unknown `res` key in AttributeType `attribute` response: {0}".format(whichone))

    def data_type(self):
        """ Get the DataType enum (grakn.DataType) corresponding to the type of this attribute """
        get_data_type_req = RequestBuilder.ConceptMethod.AttributeType.data_type()
        method_response = self._tx_service.run_concept_method(self.id, get_data_type_req)
        response = method_response.attributeType_dataType_res
        whichone = response.WhichOneof('res')
        if whichone == 'dataType':
            # iterate over enum DataType enum to find matching data type
            for e in enums.DataType:
                if e.value == response.dataType:
                    return e
            else:
                # loop exited normally
                raise GraknError("Reported datatype NOT in enum: {0}".format(response.dataType))
        elif whichone == 'null':
            return None
        else:
            raise GraknError("Unknown datatype response for AttributeType: {0}".format(whichone))

    def regex(self, pattern=None):
        """ Get or set regex """
        if pattern is None:
            get_regex_req = RequestBuilder.ConceptMethod.AttributeType.get_regex()
            method_response = self._tx_service.run_concept_method(self.id, get_regex_req)
            return method_response.attributeType_getRegex_res.regex
        else:
            set_regex_req = RequestBuilder.ConceptMethod.AttributeType.set_regex(pattern)
            method_response = self._tx_service.run_concept_method(self.id, set_regex_req)
            return self
    regex.__annotations__ = {'pattern': str}


class RelationType(Type):
    def create(self):
        """ Create an instance of a relation with this type """
        create_rel_inst_req = RequestBuilder.ConceptMethod.RelationType.create()
        method_response = self._tx_service.run_concept_method(self.id, create_rel_inst_req)
        grpc_relation_concept = method_response.relationType_create_res.relation
        from grakn.service.Session.Concept import ConceptFactory
        return ConceptFactory.create_concept(self._tx_service, grpc_relation_concept)
        
    def roles(self):
        """ Retrieve roles in this relation schema type """
        get_roles = RequestBuilder.ConceptMethod.RelationType.roles()
        method_response = self._tx_service.run_concept_method(self.id, get_roles)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.relationType_roles_iter.id,
                lambda tx_serv, iter_res:
                    ConceptFactory.create_concept(tx_serv,
                    iter_res.conceptMethod_iter_res.relationType_roles_iter_res.role)
                )
        

    def relates(self, role):
        """ Set a role in this relation schema type """
        relates_req = RequestBuilder.ConceptMethod.RelationType.relates(role)
        method_response = self._tx_service.run_concept_method(self.id, relates_req)
        return self       

    def unrelate(self, role):
        """ Remove a role in this relation schema type """
        unrelate_req = RequestBuilder.ConceptMethod.RelationType.unrelate(role)
        method_response = self._tx_service.run_concept_method(self.id, unrelate_req)
        return self

class Rule(SchemaConcept):

    def get_when(self):
        """ Retrieve the `when` clause for this rule """
        when_req = RequestBuilder.ConceptMethod.Rule.when()
        method_response = self._tx_service.run_concept_method(self.id, when_req)
        response = method_response.rule_when_res
        whichone = response.WhichOneof('res')
        if whichone == 'pattern':
            return response.pattern
        elif whichone == 'null':
            return None
        else:
            raise GraknError("Unknown field in get_when of `rule`: {0}".format(whichone))

    def get_then(self):
        """ Retrieve the `then` clause for this rule """
        then_req = RequestBuilder.ConceptMethod.Rule.then()
        method_response = self._tx_service.run_concept_method(self.id, then_req)
        response = method_response.rule_then_res
        whichone = response.WhichOneof('res')
        if whichone == 'pattern':
            return response.pattern
        elif whichone == 'null':
            return None
        else:
            raise GraknError("Unknown field in get_then or `rule`: {0}".format(whichone))

class Role(SchemaConcept):

    def relations(self):
        """ Retrieve relations that this role participates in, as an iterator """
        relations_req = RequestBuilder.ConceptMethod.Role.relations()
        method_response = self._tx_service.run_concept_method(self.id, relations_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.role_relations_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.role_relations_iter_res.relationType)
               )

    def players(self):
        """ Retrieve an iterator of entities that play this role """
        players_req = RequestBuilder.ConceptMethod.Role.players()
        method_response = self._tx_service.run_concept_method(self.id, players_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.role_players_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.role_players_iter_res.type)
               )


class Thing(Concept):

    def is_inferred(self):
        """ Is this instance inferred """
        is_inferred_req = RequestBuilder.ConceptMethod.Thing.is_inferred()
        method_response = self._tx_service.run_concept_method(self.id, is_inferred_req)
        return method_response.thing_isInferred_res.inferred
    is_inferred.__annotations__ = {'return': bool}

    def type(self):
        """ Get the type (schema concept) of this Thing """
        type_req = RequestBuilder.ConceptMethod.Thing.type()
        method_response = self._tx_service.run_concept_method(self.id, type_req)
        from grakn.service.Session.Concept import ConceptFactory
        return ConceptFactory.create_concept(self._tx_service, method_response.thing_type_res.type)

    def relations(self, *roles):
        """ Get iterator this Thing's relations, filtered to the optionally provided roles """
        relations_req = RequestBuilder.ConceptMethod.Thing.relations(roles)
        method_response = self._tx_service.run_concept_method(self.id, relations_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.thing_relations_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.thing_relations_iter_res.relation)
               )

    def attributes(self, *attribute_types):
        """ Retrieve iterator of this Thing's attributes, filtered by optionally provided attribute types """
        attrs_req = RequestBuilder.ConceptMethod.Thing.attributes(attribute_types)
        method_response = self._tx_service.run_concept_method(self.id, attrs_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.thing_attributes_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.thing_attributes_iter_res.attribute)
               )

    def roles(self):
        """ Retrieve iterator of roles this Thing plays """
        roles_req = RequestBuilder.ConceptMethod.Thing.roles()
        method_response = self._tx_service.run_concept_method(self.id, roles_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.thing_roles_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.thing_roles_iter_res.role)
               )

    def keys(self, *attribute_types):
        """ Retrieve iterator of keys (i.e. actual attributes) of this Thing, filtered by the optionally provided attribute types """
        keys_req = RequestBuilder.ConceptMethod.Thing.keys(attribute_types)
        method_response = self._tx_service.run_concept_method(self.id, keys_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.thing_keys_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.thing_keys_iter_res.attribute)
               )


    def has(self, attribute):
        """ Attach an attribute instance to this Thing """
        has_req = RequestBuilder.ConceptMethod.Thing.has(attribute)
        method_response = self._tx_service.run_concept_method(self.id, has_req)
        return


    def unhas(self, attribute):
        """ Remove an attribute instance from this Thing """
        unhas_req = RequestBuilder.ConceptMethod.Thing.unhas(attribute)
        method_response = self._tx_service.run_concept_method(self.id, unhas_req)
        return 


class Entity(Thing):
    pass

class Attribute(Thing):

    def value(self):
        """ Retrieve the value contained in this Attribute instance """
        value_req = RequestBuilder.ConceptMethod.Attribute.value()
        method_response = self._tx_service.run_concept_method(self.id, value_req)
        grpc_value_object = method_response.attribute_value_res.value
        from grakn.service.Session.util import ResponseReader
        return ResponseReader.ResponseReader.from_grpc_value_object(grpc_value_object)

    def owners(self):
        """ Retrieve entities that have this attribute value """
        owners_req = RequestBuilder.ConceptMethod.Attribute.owners()
        method_response = self._tx_service.run_concept_method(self.id, owners_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.attribute_owners_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.attribute_owners_iter_res.thing)
               )


class Relation(Thing):

    def role_players_map(self):
        """ Retrieve dictionary {role : set(players)} for this relation """
        role_players_map_req = RequestBuilder.ConceptMethod.Relation.role_players_map()
        method_response = self._tx_service.run_concept_method(self.id, role_players_map_req)

        # create the iterator to obtain all the pairs of (role, player)
        def to_pair(tx_service, iter_res):
            response = iter_res.conceptMethod_iter_res.relation_rolePlayersMap_iter_res
            from grakn.service.Session.Concept import ConceptFactory
            role = ConceptFactory.create_concept(tx_service, response.role)
            from grakn.service.Session.Concept import ConceptFactory
            player = ConceptFactory.create_concept(tx_service, response.player)
            return (role, player)

        from grakn.service.Session.util import ResponseReader
        iterator = ResponseReader.ResponseReader.iter_res_to_iterator(
                    self._tx_service,
                    method_response.relation_rolePlayersMap_iter.id,
                    to_pair)
        
        # collect all pairs of (role, player) from the iterator (executes over network to Grakn server)
        pairs = list(iterator)
        
        # aggregate into a map from role to set(player)
        # note: need to use role ID as the map key ultimately
        mapping = {}
        id_mapping = {}
        for (role, player) in pairs:
            role_id = role.id
            if role_id in id_mapping:
                role_key = id_mapping[role_id]
            else:
                id_mapping[role_id] = role
                role_key = role
                mapping[role_key] = set()
            mapping[role_key].add(player)

        return mapping

    def role_players(self, *roles):
        """ Retrieve role players filtered by roles """
        role_players_req = RequestBuilder.ConceptMethod.Relation.role_players(roles)
        method_response = self._tx_service.run_concept_method(self.id, role_players_req)
        from grakn.service.Session.util import ResponseReader
        from grakn.service.Session.Concept import ConceptFactory
        return ResponseReader.ResponseReader.iter_res_to_iterator(
                self._tx_service,
                method_response.relation_rolePlayers_iter.id,
                lambda tx_service, iter_res:
                    ConceptFactory.create_concept(tx_service, iter_res.conceptMethod_iter_res.relation_rolePlayers_iter_res.thing)
               )


    def assign(self, role, thing):
        """ Assign an entity to a role on this relation instance """
        assign_req = RequestBuilder.ConceptMethod.Relation.assign(role, thing)
        method_response = self._tx_service.run_concept_method(self.id, assign_req)
        return self

    def unassign(self, role, thing):
        """ Un-assign an entity from a role on this relation instance """
        unassign_req = RequestBuilder.ConceptMethod.Relation.unassign(role, thing)
        method_response = self._tx_service.run_concept_method(self.id, unassign_req)
        return self



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from grakn_protocol.keyspace import Keyspace_pb2 as keyspace_dot_Keyspace__pb2


class KeyspaceServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.retrieve = channel.unary_unary(
        '/keyspace.KeyspaceService/retrieve',
        request_serializer=keyspace_dot_Keyspace__pb2.Keyspace.Retrieve.Req.SerializeToString,
        response_deserializer=keyspace_dot_Keyspace__pb2.Keyspace.Retrieve.Res.FromString,
        )
    self.delete = channel.unary_unary(
        '/keyspace.KeyspaceService/delete',
        request_serializer=keyspace_dot_Keyspace__pb2.Keyspace.Delete.Req.SerializeToString,
        response_deserializer=keyspace_dot_Keyspace__pb2.Keyspace.Delete.Res.FromString,
        )


class KeyspaceServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def retrieve(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def delete(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_KeyspaceServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'retrieve': grpc.unary_unary_rpc_method_handler(
          servicer.retrieve,
          request_deserializer=keyspace_dot_Keyspace__pb2.Keyspace.Retrieve.Req.FromString,
          response_serializer=keyspace_dot_Keyspace__pb2.Keyspace.Retrieve.Res.SerializeToString,
      ),
      'delete': grpc.unary_unary_rpc_method_handler(
          servicer.delete,
          request_deserializer=keyspace_dot_Keyspace__pb2.Keyspace.Delete.Req.FromString,
          response_serializer=keyspace_dot_Keyspace__pb2.Keyspace.Delete.Res.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'keyspace.KeyspaceService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import sonnet as snt
import tensorflow as tf


class TypewiseEncoder(snt.AbstractModule):
    """
    Orchestrates encoding elements according to their types. Defers encoding of each feature to the appropriate encoder
    for the type of that feature. Assumes that the type is given categorically as an integer value in the 0th position
    of the provided features Tensor.
    """
    def __init__(self, encoders_for_types, feature_length, name="typewise_encoder"):
        """
        Args:
            encoders_for_types: Dict - keys: encoders; values: a list of type categories the encoder should be used for
            feature_length: The length of features to output for matrix initialisation
            name: The name for this Module
        """
        super(TypewiseEncoder, self).__init__(name=name)

        types_considered = []
        for a in encoders_for_types.values():
            types_considered.extend(a)
        types_considered.sort()

        expected_types = list(range(max(types_considered) + 1))

        if types_considered != expected_types:
            raise ValueError(
                f'Encoder categories are inconsistent. Expected {expected_types}, but got {types_considered}')

        self._feature_length = feature_length
        self._encoders_for_types = encoders_for_types

    def _build(self, features):

        tf.summary.histogram('typewise_encoder_features_histogram', features)

        shape = tf.stack([tf.shape(features)[0], self._feature_length])

        encoded_features = tf.zeros(shape, dtype=tf.float32)

        for encoder, types in self._encoders_for_types.items():

            feat_types = tf.cast(features[:, 0], tf.int32)  # The types for each feature, as integers

            # Expand dimensions ready for element-wise equality comparison
            exp_types = tf.expand_dims(types, axis=0)
            exp_feat_types = tf.expand_dims(feat_types, axis=1)

            elementwise_equality = tf.equal(exp_feat_types, exp_types)

            # Use this encoder when the feat_type matches any of the types
            applicable_types_mask = tf.reduce_any(elementwise_equality, axis=1)
            indices_to_encode = tf.where(applicable_types_mask)

            feats_to_encode = tf.squeeze(tf.gather(features[:, 1:], indices_to_encode), axis=1)
            encoded_feats = encoder()(feats_to_encode)

            encoded_features += tf.scatter_nd(tf.cast(indices_to_encode, dtype=tf.int32), encoded_feats, shape)

        tf.summary.histogram('typewise_encoder_encoded_features_histogram', encoded_features)

        return encoded_features



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from grakn_protocol.session import Session_pb2 as session_dot_Session__pb2


class SessionServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.open = channel.unary_unary(
        '/session.SessionService/open',
        request_serializer=session_dot_Session__pb2.Session.Open.Req.SerializeToString,
        response_deserializer=session_dot_Session__pb2.Session.Open.Res.FromString,
        )
    self.close = channel.unary_unary(
        '/session.SessionService/close',
        request_serializer=session_dot_Session__pb2.Session.Close.Req.SerializeToString,
        response_deserializer=session_dot_Session__pb2.Session.Close.Res.FromString,
        )
    self.transaction = channel.stream_stream(
        '/session.SessionService/transaction',
        request_serializer=session_dot_Session__pb2.Transaction.Req.SerializeToString,
        response_deserializer=session_dot_Session__pb2.Transaction.Res.FromString,
        )


class SessionServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def open(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def close(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def transaction(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_SessionServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'open': grpc.unary_unary_rpc_method_handler(
          servicer.open,
          request_deserializer=session_dot_Session__pb2.Session.Open.Req.FromString,
          response_serializer=session_dot_Session__pb2.Session.Open.Res.SerializeToString,
      ),
      'close': grpc.unary_unary_rpc_method_handler(
          servicer.close,
          request_deserializer=session_dot_Session__pb2.Session.Close.Req.FromString,
          response_serializer=session_dot_Session__pb2.Session.Close.Res.SerializeToString,
      ),
      'transaction': grpc.stream_stream_rpc_method_handler(
          servicer.transaction,
          request_deserializer=session_dot_Session__pb2.Transaction.Req.FromString,
          response_serializer=session_dot_Session__pb2.Transaction.Res.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'session.SessionService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from grakn_protocol.keyspace.Keyspace_pb2_grpc import KeyspaceServiceStub
import grakn_protocol.keyspace.Keyspace_pb2 as keyspace_messages


class KeyspaceService(object):

    def __init__(self, uri, channel, credentials=None):
        self.uri = uri
        self.stub = KeyspaceServiceStub(channel)
        self.credentials = credentials

    def retrieve(self):
        retrieve_request = keyspace_messages.Keyspace.Retrieve.Req()
        if self.credentials:
            retrieve_request.username = self.credentials['username']
            retrieve_request.password = self.credentials['password']
        response = self.stub.retrieve(retrieve_request)
        return list(response.names)

    def delete(self, keyspace):
        delete_request = keyspace_messages.Keyspace.Delete.Req()
        delete_request.name = keyspace
        if self.credentials:
            delete_request.username = self.credentials['username']
            delete_request.password = self.credentials['password']
        self.stub.delete(delete_request)
        return



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import os
import shutil
import subprocess as sp
import tempfile
import zipfile


class ZipFile(zipfile.ZipFile):
    """ Custom ZipFile class handling file permissions. """
    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        targetpath = super()._extract_member(member, targetpath, pwd)

        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(targetpath, attr)
        return targetpath


class GraknServer(object):
    DISTRIBUTION_LOCATION = 'external/graknlabs_grakn_core/grakn-core-all-mac.zip'
    DISTRIBUTION_ROOT_DIR = 'grakn-core-all-mac'

    def __init__(self):
        self.__unpacked_dir = None

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.stop()

    def start(self):
        if not self.__unpacked_dir:
            self._unpack()
        sp.check_call([
            'grakn', 'server', 'start'
        ], cwd=os.path.join(self.__unpacked_dir, GraknServer.DISTRIBUTION_ROOT_DIR))
        return self

    def stop(self):
        sp.check_call([
            'grakn', 'server', 'stop'
        ], cwd=os.path.join(self.__unpacked_dir, GraknServer.DISTRIBUTION_ROOT_DIR))
        shutil.rmtree(self.__unpacked_dir)
        return self

    def _unpack(self):
        self.__unpacked_dir = tempfile.mkdtemp(prefix='grakn')
        with ZipFile(GraknServer.DISTRIBUTION_LOCATION) as zf:
            zf.extractall(self.__unpacked_dir)

    def load_graql_file(self, keyspace, graql_file_path):
        sp.check_call([
            'grakn', 'console', '-k', keyspace, '-f',
            os.getenv("TEST_SRCDIR") + graql_file_path
        ], cwd=self.grakn_binary_location)

    @property
    def grakn_binary_location(self):
        return os.path.join(self.__unpacked_dir, GraknServer.DISTRIBUTION_ROOT_DIR)



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#


def get_thing_types(tx):
    """
    Get all schema types, excluding those for implicit attribute relations and base types
    Args:
        tx: Grakn transaction

    Returns:
        Grakn types
    """
    schema_concepts = tx.query(
        "match $x sub thing; "
        "not {$x sub @has-attribute;}; "
        "not {$x sub @key-attribute;}; "
        "get;")
    thing_types = [schema_concept.get('x').label() for schema_concept in schema_concepts]
    [thing_types.remove(el) for el in ['thing', 'relation', 'entity', 'attribute']]
    return thing_types


def get_role_types(tx):
    """
    Get all schema roles, excluding those for implicit attribute relations, the base role type
    Args:
        tx: Grakn transaction

    Returns:
        Grakn roles
    """
    schema_concepts = tx.query(
        "match $x sub role; "
        "not{$x sub @key-attribute-value;}; "
        "not{$x sub @key-attribute-owner;}; "
        "not{$x sub @has-attribute-value;}; "
        "not{$x sub @has-attribute-owner;};"
        "get;")
    role_types = ['has'] + [role.get('x').label() for role in schema_concepts]
    role_types.remove('role')
    return role_types



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

from kglib.utils.grakn.object.comparable import PropertyComparable

DATA_TYPE_NAMES = ('long', 'double', 'boolean', 'date', 'string')


class Thing(PropertyComparable):
    def __init__(self, id, type_label, base_type_label, data_type=None, value=None):
        self.id = id
        self.type_label = type_label
        self.base_type_label = base_type_label  # TODO rename to base_type in line with Client Python

        # If the thing is an attribute
        self.data_type = data_type
        self.value = value

        # TODO Make attribute a separate class
        if self.base_type_label == 'attribute':
            if self.data_type is None:
                raise ValueError('Attribute data_type must be provided')
            if self.value is None:
                raise ValueError('Attribute value must be provided')

    def __str__(self):
        string = f'<{self.type_label}, {self.id}'
        if self.base_type_label == 'attribute':
            string += f': {self.value}'
        return string + '>'

    def __repr__(self):
        return self.__str__()


def build_thing(grakn_thing):

    id = grakn_thing.id
    type_label = grakn_thing.type().label()
    base_type_label = grakn_thing.base_type.lower()

    assert(base_type_label in ['entity', 'relation', 'attribute'])

    if base_type_label == 'attribute':
        data_type = grakn_thing.type().data_type().name.lower()
        assert data_type in DATA_TYPE_NAMES
        value = grakn_thing.value()

        return Thing(id, type_label, base_type_label, data_type, value)

    return Thing(id, type_label, base_type_label)
